/**
 * Benchmark daemon for RITAS.
 */
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include "benchmark.h"

void usage(char *argv) {
	printf("Usage: %s <pid> <port> <procfile>\n", argv);
}

int main(int argc, char **argv) {
	
	int n, i;
	struct sockaddr_in addr, peeraddr;
	socklen_t addr_s;
	int sock, tcpsock;
	u_short pid, port;
	char *s, buf[BUF_S], buf_port[8];
	u_char protocol;
	FILE *fp;
	ritas_t *ctx;
	char errbuf[RITAS_ERRBUF_SIZE];
	u_char key[RITAS_KEY_SIZE];
	u_short procn;
	
	/*
	 * Parse arguments;
	 */
	if(argc < 3) {
		usage(argv[0]);
		exit(1);
	}
	
	setlinebuf(stdout);
	
	pid = atoi(argv[1]);
	port = atoi(argv[2]);

	/*
	 * Open and parse processes' file
	 */
	fp = fopen(argv[3], "r");
	if(!fp) {
		perror("fopen");
		exit(1);
	}

	s = fgets(buf, BUF_S, fp);
	if(!s) {
		perror("fgets");
		exit(1);
	}

	procn = atoi(buf);
	if(procn < 4 || procn > 255) {
		printf("invalid number of processes\n");
	}

	/*
	 * Initialize RITAS context.
	 */
	ctx = ritas_init(pid, procn, (procn-1)/3, port, errbuf);
	if(!ctx) {
		printf("%s\n", errbuf);
		exit(1);
	}

	memset(key, '\0', RITAS_KEY_SIZE);
		
	for(i = 0 ; i < procn ; i++) {
		s = fgets(buf, BUF_S, fp);
		if(!s) {
			perror("fgets");
			exit(1);
		}
		buf[strlen(buf)-1] = '\0';
		
		s = fgets(buf_port, 8, fp);
		if(!s) {
			perror("fgets");
			exit(1);
		}
		port = atoi(buf_port);

		ritas_proc_add_ipv4(ctx, i, (u_char *)buf, port, key);
	}
	fclose(fp);
	
	sleep(2);

	n = ritas_start(ctx);
	if(n < 0) {
		printf("%s\n", ctx->errbuf);
		exit(1);
	}
	
	/*
	 * Create socket.
	 */
	sock = socket(AF_INET, SOCK_DGRAM, 0);
	if(sock < 0) {
		perror("socket");
		exit(1);
	}

	memset(&addr, '\0', sizeof(addr));
	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = htonl(INADDR_ANY);
	addr.sin_port = htons(BASE_PORT+pid);
	
	n = bind(sock, (struct sockaddr *)&addr, sizeof(addr));
	if(n < 0) {
		perror("bind");
		exit(1);
	}

	addr_s = sizeof(peeraddr);

	n = recvfrom(sock, buf, BUF_S, 0, (struct sockaddr *)&peeraddr, &addr_s);
	if(n < 0) {
		perror("recvfrom");
		exit(1);
	}
	
	protocol = *buf;
	
	switch(protocol) {
		case RB:
		rb_bench(ctx, (struct rb_bench *)buf);
		break;

		case BC:
		//bc_bench(ctx, (struct bc_bench *)buf);
		bc_bench_multi(ctx, (struct bc_bench *)buf);
		break;
		
		case MVC:
		mvc_bench(ctx, (struct mvc_bench *)buf);
		break;
		
		case VC:
		vc_bench(ctx, (struct vc_bench *)buf);
		break;
		
		case AB:
		ab_bench(ctx, (struct ab_bench *)buf);
		break;
		
		default:
		printf("invalid protocol\n");
		exit(1);
	}
	

	//ritas_destroy(ctx);
	sendto(sock, buf, 1, 0, (struct sockaddr *)&peeraddr, sizeof(peeraddr));
	if(n < 0) {
		perror("sendto");
		exit(1);
	}

/*	
	tcpsock = socket(AF_INET, SOCK_STREAM, 0);
	peeraddr.sin_port = htons(BASE_PORT);
	connect(tcpsock, (struct sockaddr *)&peeraddr, sizeof(peeraddr));
	close(tcpsock);
*/	
	sleep(2);

	exit(0);			
}

int rb_bench(ritas_t *ctx, struct rb_bench *msg) {
	
	u_char *buf;
	int n;
	struct timeval t0, t1;
	long long dt;

	buf = malloc(msg->value_s);
	
	if(msg->sender == ctx->pid) {
		memset(buf, 'A', msg->value_s);
		gettimeofday(&t0, NULL);
		n = ritas_rb_bcast(ctx, msg->id, buf, msg->value_s);
		gettimeofday(&t1, NULL);
		if(n == -1) {
			printf("%s\n", ctx->errbuf);
			exit(1);
		}

		dt = toLL(t1)-toLL(t0);
		printf("[%d] RB %d, delivery time: %lld us\n", ctx->pid, msg->id, dt);
	}
	else {
		n = ritas_rb_recv(ctx, msg->sender, msg->id, buf, msg->value_s);	
		if(n == -1) {
			printf("%s\n", ctx->errbuf);
			exit(1);
		}
	}
	
	free(buf);
	
	return 1;
}

int bc_bench(ritas_t *ctx, struct bc_bench *msg) {
	
	int n;
	struct timeval t0, t1;
	long long dt;
	u_char proposal;
	
	//proposal = 1;
	proposal = ctx->pid % 2;
	gettimeofday(&t0, NULL);
	ritas_reset_timer(ctx);

	n = ritas_bc(ctx, msg->id, proposal);
	gettimeofday(&t1, NULL);
	if(n == -1) {
		printf("%s\n", ctx->errbuf);
		exit(1);
	}

	dt = toLL(t1)-toLL(t0);
	printf("[%d] BC %d, decision time: %lld us, value: %d\n", ctx->pid, msg->id, dt, n);

	return 1;
}

int bc_bench_multi(ritas_t *ctx, struct bc_bench *msg) {
	
	int n;
	u_char proposal;

	proposal = 1;	
	#ifdef PD_UNIFORM
	proposal = 1;
	#endif
	#ifdef PD_CORROSIVE
	proposal = ctx->pid % 2;
	#endif

	ritas_bc_multi(ctx, msg->id, proposal, msg->requests);
	
	pthread_mutex_lock(&ctx->pr_mutex);
	while(ctx->pr_count)
		pthread_cond_wait(&ctx->pr_cond, &ctx->pr_mutex);
	pthread_mutex_unlock(&ctx->pr_mutex);
	
	printf("BC BURST END %d\n", msg->requests);
	return 1;
}

int mvc_bench(ritas_t *ctx, struct mvc_bench *msg) {
	
	u_char *buf, *decision;
	u_short decision_s;
	int n;
	struct timeval t0, t1;
	long long dt;

	buf = malloc(msg->value_s);
	#ifndef PD_CORROSIVE
	memset(buf, 'A', msg->value_s);
	#else
	memset(buf, 'A'+ctx->pid, msg->value_s);
	#endif

	decision = malloc(msg->value_s);
	decision_s = msg->value_s;

	gettimeofday(&t0, NULL);
	ritas_reset_timer(ctx);
	n = ritas_mvc(ctx, msg->id, buf, msg->value_s, decision, decision_s);
	gettimeofday(&t1, NULL);
	if(n == -1) {
		printf("%s\n", ctx->errbuf);
		exit(1);
	}

	free(buf);

	dt = toLL(t1)-toLL(t0);
	
	if(n == 0)
		printf("[%d] MVC %d, decision time: %lld us, default value\n", ctx->pid, msg->id, dt);		
	else
		printf("[%d] MVC %d, decision time: %lld us, first byte: 0x%x (%d bytes)\n", ctx->pid, msg->id, dt, decision[0], n);

	free(decision);
	
	sleep(2);
	return 1;
}

int vc_bench(ritas_t *ctx, struct vc_bench *msg) {

	u_char *buf;
	u_char decision[256];
	ritas_vector_t *vector;
	int n, i;
	struct timeval t0, t1;
	long long dt;

	buf = malloc(msg->value_s);
	memset(buf, 'A'+ctx->pid, msg->value_s);
	
	gettimeofday(&t0, NULL);
	ritas_reset_timer(ctx);
	n = ritas_vc(ctx, msg->id, buf, msg->value_s, decision, 256);
	gettimeofday(&t1, NULL);
	if(n == -1) {
		printf("%s\n", ctx->errbuf);
		exit(1);
	}

	free(buf);

	dt = toLL(t1)-toLL(t0);

	vector = ritas_vector_extract(decision, 256);
	if(!vector) {
		printf("vector extraction error\n");
		exit(1);
	}
		
	printf("[%d] VC %d, decision time: %lld us, %d elements\n", ctx->pid, msg->id, dt, vector->size);

	for(i = 0 ; i < vector->size ; i++)
		printf("[%d] VC %d, decision vector[%d]: 0x%x (%d bytes)\n", ctx->pid, msg->id, i, (vector->element[i] ? *vector->element[i] : 0), vector->element_s[i]);

	ritas_vector_destroy(vector);
	return 1;
}

int ab_bench(ritas_t *ctx, struct ab_bench *msg) {
	
	int i, n, mult;
	u_char *buf;
	struct timeval *t0, t1, bt0, bt1;
	long long dt;
	struct ctxmsg cm;
	ritas_ab_header_t abh;
	pthread_t tx_thread;
	
	n = ritas_ab_init(ctx);
	if(n == -1) {
		printf("%s\n", ctx->errbuf);
		exit(1);
	}
	
	sleep(2);
	
	printf("BEGIN AB MEASUREMENT, burst = %d, m = %d, sender = %d\n", msg->requests, msg->value_s, msg->sender);
	gettimeofday(&bt0, NULL);
	gettimeofday(&ctx->t_init, NULL);

	if(msg->sender == ctx->pid || msg->sender == -1) {
		cm.ctx = ctx;
		cm.msg = msg;
		n = pthread_create(&tx_thread, NULL, (void *(*)(void *))&ab_tx, (void *)&cm);
		if(n) {
			printf("error creating thread\n");
			exit(1);
		}
	}	
	
	buf = malloc(msg->value_s);
	if(msg->sender == -1)
		mult = ctx->n;
	else
		mult = 1;

	cm.ctx = ctx;
	cm.msg = msg;		

	for(i = 0 ; i < mult*msg->requests ; i++) {
		n = ritas_ab_recv(ctx, buf, msg->value_s, &abh);
		//gettimeofday(&t1, NULL);
		if(n == -1) {
			printf("error receiving message\n");
			exit(1);
		}
		/*
		if(msg->sender == ctx->pid || (msg->sender == -1 && abh.txid == ctx->pid)) {
			t0 = (struct timeval *)(buf);
			dt = toLL(t1)-toLL((*t0));
			abprintf("[%d] AB request time: %lld us\n", ctx->pid, dt);
			abprintf("[%d] AB msg from %d, value: %c* (%d bytes)\n", ctx->pid, abh.txid, *(buf+8), n);
		}
		*/
	}
		
	gettimeofday(&bt1, NULL);
	free(buf);

	dt = toLL(bt1)-toLL(bt0);
	printf("END AB MEASUREMENT, burst delivery time: %lld us (%lld us/req), %d consensus\n", dt, dt/msg->requests, (ctx->abcb->vcid-1));
	
	if(msg->sender == ctx->pid || msg->sender == -1) {
		n = pthread_join(tx_thread, NULL);
		if(n) {
			printf("error joining thread\n");
			exit(1);
		}
	}
	
	sleep(3);
	ritas_ab_destroy(ctx);
	
	return 1;
}


void ab_tx(struct ctxmsg *cm) {
	u_char *buf;
	int n, i;
	struct timeval *t0, to, tv;

	tv.tv_sec = cm->msg->rate/1000;
	tv.tv_usec = (cm->msg->rate - cm->msg->rate/1000*1000)*1000;

	buf = malloc(cm->msg->value_s);
	memset(buf, 'A'+cm->ctx->pid, cm->msg->value_s);

	for(i = 0 ; i < cm->msg->requests ; i++) {
		t0 = (struct timeval *)buf;
		gettimeofday(t0, NULL);
		n = ritas_ab_bcast(cm->ctx, buf, cm->msg->value_s);
		if(n == -1) {
			printf("error broadcasting message\n");
			exit(1);
		}		
		
		to.tv_sec = tv.tv_sec;
		to.tv_usec = tv.tv_usec;

		dprintf("BENCHD: broadcast time: %d.%d\n", t0->tv_sec, t0->tv_usec);
		dprintf("BENCHD: WAITING FOR %d s AND %d us\n", to.tv_sec, to.tv_usec);
		
		n = select(0, NULL, NULL, NULL, &to);
		if(n == -1) {
			printf("error in select()\n");
			exit(1);
		}		
	}
	
	free(buf);
	
	pthread_exit(NULL);
}
