#include "libdit-structures.h"


/*
 * Group management functions.
 */

/**
 * Adds a process to the communication group.
 * The process is referenced by a tuple consisting of: a string containing one 
 * of the following:
 *   - the IPv4 dotted-decimal address
 *   - the IPv6 hex string address
 * and a port number.
 * @param ctx libdit descriptor.
 * @param addr the IPv4 dotted-decimal or the IPv6 hex string format.
 * @return a pointer to the process upon success; NULL upon failure.
 */ 
libdit_proc_t *libdit_group_add_ip(libdit_t *ctx, u_char *addr, u_short port);

/**
 * Adds a process to the communication group.
 * The process is referenced by a sockaddr structure.
 * @param ctx libdit descriptor.
 * @param addr sockaddr structure containing the address of the process.
 * @return a pointer to the process upon success; NULL upon failure.
 */
libdit_proc_t *libdit_group_add_sockaddr(libdit_t *ctx, struct sockaddr *addr);

/**
 * Removes a process from the communication group.
 * @param ctx libdit descriptor.
 * @param proc pointer to the process to be removed.
 * @return 1 upon success; 0 upon failure.
 */
int libdit_group_remove(libdit_t *ctx, libdit_proc_t *proc);

/**
 * Returns the size of the communication group.
 * @param ctx libdit descriptor
 * @return the size of the group.
 */ 
int libdit_group_size(libdit_t *ctx);




/*
 * I/O primitives
 */
 
/**
 * Broadcasts a message to all the processes in a group. The contents of 
 * <code>buf</code> are sent to all the process in the communication group 
 * refered by the libdit descriptor <code>ctx</code>. The variable 
 * <code>buf_s</code> specifies the length of <code>buf</code> in
 * bytes. The function is blocking.
 * @param ctx libdit descriptor.
 * @param buf pointer to a buffer containing the message to be transmitted.
 * @param buf_s buffer length in bytes.
 * @return 1 upon success; 0 upon failure.
 */
int libdit_broadcast(libdit_t *ctx, char buf, u_int buf_s);

/**
 * Sends a message to an arbitraty process in a group. The contents of 
 * <code>buf</code> are sent to a single process, specified by the pointer 
 * <code>proc</proc>, which belongs to the communication group refered by the
 * libdit descriptor <code>ctx</code>. The variable <code>buf_s</code> 
 * specifies the length of <code>buf</code> in bytes. The function is blocking.
 * @param ctx libdit descriptor.
 * @param proc process to which the message is destined.
 * @param buf pointer to a buffer containing the message to be transmitted.
 * @param buf_s buffer length in bytes.
 * @return the number of bytes transmitted upon success; -1 upon failure.
 */
int libdit_send(libdit_t *ctx, libdit_proc_t *proc, char buf, u_int buf_s);

/**
 * Receives a message from any given process in a group. The buffer 
 * <code>buf</buf> is filled up to <code>buf_s</code> bytes with a message 
 * received by the first process who sent it. This routine assures messages
 * are received on a first-come, first-served basis.
 * @param ctx libdit descriptor.
 * @param proc upon return holds a pointer to the process sending the message.
 * @param buf pointer to a buffer which will hold the message received.
 * @param buf_s buffer length in bytes.
 * @return the number of bytes received upon success; -1 upon failure.
 */
int libdit_recv(libdit_t *ctx, libdit_proc &proc, char buf, u_int buf_s);

/**
 * Initializes a new value list.
 */
libdit_vlist_t *libdit_vlist_init();

/**
 * Destroys a value list.
 */
void libdit_vlist_destroy(libdit_vlist_t *vlist);

/**
 * Stores a value in a value list.
 */
int libdit_vlist_store(libdit_vlist_t *vlist, char *value, u_int value_s);

/**
 * Returns the number of ocurrences of a value in a value list.
 */
int libdit_vlist_count(libdit_vlist_t *vlist, char *value, u_int value_s);

/**
 * Builds a Reliable Broadcast message.
 */
int libdit_rb_build(char *buf, u_int buf_s, int type, char *value, u_int value_s);

/**
 * Extracts a Reliable Broadcast message.
 */
int libdit_rb_extract(libdit_rb_msg_t *msg, char *buf, u_int buf_s);


/**
 * Initializes a new communication session context.
 * @see libdit_destroy()
 * @param ab_type the atomic broadcast protocol.
 * @param vc_type the vector consensus protocol.
 * @param mvc_type the multivalued consensus protocol.
 * @param bc_type the binary consensus protocol.
 * @param rb_type the reliable broadcast protocol.
 * @param rc_type the reliable channel type.
 * @return a libdit descriptor upon sucess; NULL upon failure.
 */ 
libdit_t *libdit_init(int ab_type, int vc_type, int mvc_type, bc_type,
	rb_type, rc_type);

/**
 * Closes a communication session context.
 * Performs cleanup functions associated with a context. This
 * function does not fail.
 * @see libdit_init()
 * @param ctx the libdit descriptor referencing the session to be free'd.
 */ 
void libdit_destroy(libdit_t *ctx);


/*
 * Performs an atomic broadcast. The values on the parameters
 * are role-dependent. If sender they carry the message to be atomically
 * broadcasted; if receiver they reference the buffer to receive the message.
 * @param ctx the libdit descriptor.
 * @param buf the message to be transmitted.
 * @param buf_s the size of the message (in bytes).
 * @return 1 upon success; 0 upon failure.
 */
int	libdit_atomic_broadcast(libdit_t *ctx, void *buf, u_int buf_s);
	
int	libdit_vector_consensus(libdit_t *ctx, void *buf, u_int buf_s);
	
int	libdit_multivalued_consensus(libdit_t *ctx, void *buf, u_int buf_s);

int	libdit_binary_consensus(libdit_t *ctx, u_short v);


/**
 * Performs a reliable broadcast operation.
 * There are two possible roles when calling this function:
 *   - Sender (LIBDIT_RB_SEND). We are broadcasting the message.
 *   - Receiver (LIBDIT_RB_RECV). We receiving the message.
 * 
 * This function is synchronous, meaning that will return only upon completing
 * the operation.
 * 
 * @param ctx the libdit context.
 * @param role the role can be one of two values: LIBDIT_RB_SEND or 
 *        LIBDIT_RB_RECV.
 * @param buf if role is LIBDIT_RB_SEND, a buffer containing  the message to be
 *        broadcasted; if role is LIBDIT_RV_RECV, a buffer to write the message 
 *        to be received.
 * @param buf_s if role is LIBDIT_RB_SEND, the size of the message; if role is
 *        LIBDIT_RB_RECV, the maximum number of bytes allowed to be written to 
 *        the buffer.
 * @return 1 upon success; 0 upon failure.
 */ 	
int	libdit_reliable_broadcast(libdit_t *ctx, char role, char *buf, u_int buf_s);

/**
 * Provides statistical information about a communication session.
 * A libdit_stat structure if filled with the statistical information
 * about the given context.
 * @param ctx libdit descriptor.
 * @param stats pointer to libdit_stat structure.
 * @return 1 upon success; 0 upon failure.
 */
int libdit_stats(libdit_t *ctx, struct libdit_stat *stats);