#include "ritas-structures.h"

/**
 * Initializes a new communication session context.
 * @see ritas_destroy()
 * @param ab_type the atomic broadcast protocol.
 * @param vc_type the vector consensus protocol.
 * @param mvc_type the multivalued consensus protocol.
 * @param bc_type the binary consensus protocol.
 * @param rb_type the reliable broadcast protocol.
 * @param rc_type the reliable channel type.
 * @return a ritas descriptor upon sucess; NULL upon failure.
 */ 
ritas_t *ritas_init(int ab_type, int vc_type, int mvc_type, bc_type,
	rb_type, rc_type);

/**
 * Closes a communication session context.
 * Performs cleanup functions associated with a context. This
 * function does not fail.
 * @see ritas_init()
 * @param ctx the ritas descriptor referencing the session to be free.
 */ 
void ritas_destroy(ritas_t *ctx);

/**
 * Adds a process to the communication group.
 * The process is referenced by a tuple consisting of: a string containing one 
 * of the following:
 *   - the IPv4 dotted-decimal address
 *   - the IPv6 hex string address
 * and a port number.
 * @param ctx ritas descriptor.
 * @param addr the IPv4 dotted-decimal or the IPv6 hex string format.
 * @return a pointer to the process upon success; NULL upon failure.
 */ 
ritas_proc *ritas_group_add_ip(ritas_t *ctx, u_char *addr, u_short port);

/**
 * Adds a process to the communication group.
 * The process is referenced by a sockaddr structure.
 * @param ctx ritas descriptor.
 * @param addr sockaddr structure containing the address of the process.
 * @return a pointer to the process upon success; NULL upon failure.
 */
ritas_proc *ritas_group_add_sockaddr(ritas_t *ctx, struct sockaddr *addr);

/**
 * Removes a process from the communication group.
 * @param ctx ritas descriptor.
 * @param proc pointer to the process to be removed.
 * @return 1 upon success; 0 upon failure.
 */
int ritas_group_remove(ritas_t *ctx, ritas_proc_t *proc);

/**
 * Returns the size of the communication group.
 * @param ctx ritas descriptor
 * @return the size of the group.
 */ 
int ritas_group_size(ritas_t *ctx);


/**
 * Performs a reliable broadcast.
 * 
 * This function is synchronous, meaning that will return only upon completing
 * the operation.
 * 
 * @param ctx the ritas context.
 * @param value a buffer containing the data to be
 *        broadcasted.
 * @param the length of the data to be broadcasted.
 * @return 1 upon success; 0 upon failure.
 */ 	
// RELIABLE BROADCAST
int ritas_rb_broadcast(ritas_t *ctx, u_short bsn, char *buf, u_int buf_s);
int ritas_rb_deliver(ritas_t *ctx, u_short init_proc_id, u_short bsn, char *buf, u_int buf_s);
//int ritas_rb_gather(ritas_t *ctx, u_int bsn, u_int msg_count);

// RELIABLE CHANNEL
ritas_rc_t *ritas_rc_init(ritas_t *ctx, u_short rc_id);
int ritas_rc_destroy(ritas_rc_t *rc);
int ritas_rc_sendmsg();
int ritas_rc_recvmsg();



int ritas_rb_bcastmsg(ritas_t *r, u_char type, ritas_round *k, char *value, u_short value_s);
int ritas_rb_sendmsg(ritas_t *r, ritas_proc_t *proc, u_char type, char *value, u_short value_s);
int ritas_rb_recvmsg(ritas_t *r, ritas_proc_t *proc, ritas_rb_msg *msg);
int ritas_rb_build(char buf, u_int buf_s, u_char type, ritas_round_t *k, char *value, u_short value_s);
int ritas_rb_extract(ritas_rb_msg *msg, char *buf, u_int buf_s);

int queue_store(ritas_t *r, ritas_rb_msg *msg);
int queue_retrieve(ritas_t *r, ritas_rb_msg *msg, ritas_round *k);

int ritas_rc_send(ritas_t *r, ritas_proc *proc, char *buf, u_int buf_s);
int ritas_rc_recv(ritas_t *r, ritas_proc *proc, char *buf, u_int buf_s);

int ritas_round_scheduler_init(ritas_t *r);














/*
 * I/O primitives
 */
 
/**
 * Broadcasts a message to all the processes in a group. The contents of 
 * <code>buf</code> are sent to all the process in the communication group 
 * refered by the ritas descriptor <code>ctx</code>. The variable 
 * <code>buf_s</code> specifies the length of <code>buf</code> in
 * bytes. The function is blocking.
 * @param ctx ritas descriptor.
 * @param buf pointer to a buffer containing the message to be transmitted.
 * @param buf_s buffer length in bytes.
 * @return 1 upon success; 0 upon failure.
 */
int ritas_broadcast(ritas_t *ctx, char buf, u_int buf_s);

/**
 * Sends a message to an arbitraty process in a group. The contents of 
 * <code>buf</code> are sent to a single process, specified by the pointer 
 * <code>proc</proc>, which belongs to the communication group refered by the
 * ritas descriptor <code>ctx</code>. The variable <code>buf_s</code> 
 * specifies the length of <code>buf</code> in bytes. The function is blocking.
 * @param ctx ritas descriptor.
 * @param proc process to which the message is destined.
 * @param buf pointer to a buffer containing the message to be transmitted.
 * @param buf_s buffer length in bytes.
 * @return the number of bytes transmitted upon success; -1 upon failure.
 */
int ritas_send(ritas_t *ctx, ritas_proc_t *proc, char buf, u_int buf_s);

/**
 * Receives a message from any given process in a group. The buffer 
 * <code>buf</buf> is filled up to <code>buf_s</code> bytes with a message 
 * received by the first process who sent it. This routine assures messages
 * are received on a first-come, first-served basis.
 * @param ctx ritas descriptor.
 * @param proc upon return holds a pointer to the process sending the message.
 * @param buf pointer to a buffer which will hold the message received.
 * @param buf_s buffer length in bytes.
 * @return the number of bytes received upon success; -1 upon failure.
 */
int ritas_recv(ritas_t *ctx, ritas_proc &proc, char buf, u_int buf_s);

/**
 * Initializes a new value list.
 */
ritas_vlist_t *ritas_vlist_init();

/**
 * Destroys a value list.
 */
void ritas_vlist_destroy(ritas_vlist_t *vlist);

/**
 * Stores a value in a value list.
 */
int ritas_vlist_store(ritas_vlist_t *vlist, char *value, u_int value_s);

/**
 * Returns the number of ocurrences of a value in a value list.
 */
int ritas_vlist_count(ritas_vlist_t *vlist, char *value, u_int value_s);

/**
 * Builds a Reliable Broadcast message.
 */
int ritas_rb_build(char *buf, u_int buf_s, int type, char *value, u_int value_s);

/**
 * Extracts a Reliable Broadcast message.
 */
int ritas_rb_extract(ritas_rb_msg_t *msg, char *buf, u_int buf_s);




/**
 * Performs a reliable broadcast operation.
 * There are two possible roles when calling this function:
 *   - Sender (ritas_RB_SEND). We are broadcasting the message.
 *   - Receiver (ritas_RB_RECV). We receiving the message.
 * 
 * This function is synchronous, meaning that will return only upon completing
 * the operation.
 * 
 * @param ctx the ritas context.
 * @param role the role can be one of two values: ritas_RB_SEND or 
 *        ritas_RB_RECV.
 * @param buf if role is ritas_RB_SEND, a buffer containing  the message to be
 *        broadcasted; if role is ritas_RV_RECV, a buffer to write the message 
 *        to be received.
 * @param buf_s if role is ritas_RB_SEND, the size of the message; if role is
 *        ritas_RB_RECV, the maximum number of bytes allowed to be written to 
 *        the buffer.
 * @return 1 upon success; 0 upon failure.
 */ 	
int	ritas_reliable_broadcast(ritas_t *ctx, char role, char *buf, u_int buf_s);

/**
 * Provides statistical information about a communication session.
 * A ritas_stat structure if filled with the statistical information
 * about the given context.
 * @param ctx ritas descriptor.
 * @param stats pointer to a ritas_stat structure.
 * @return 1 upon success; 0 upon failure.
 */
int ritas_stats(ritas_t *ctx, struct ritas_stat *stats);