#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <assert.h>
#include "ritas_rc.h"
#include "ritas_rc_unack.h"
#include "ritas.h"
#include "ritas_group.h"
#include "ritas_netio.h"

ritas_rccb_t *ritas_rc_init(ritas_t *ctx, u_short rcid) {
	
	ritas_rccb_t *rccb;
	int i;
	
	/*
	 * Calculate our slot in the hash table
	 */
	i = RITAS_RC_HASH(rcid);
	
	/*
	 * Prepend this rccb in the list
	 */
	if(!ctx->rccb[i]) {
		ctx->rccb[i] = (ritas_rccb_t *)malloc(sizeof(ritas_rccb_t));
		rccb = ctx->rccb[i];
		rccb->next = NULL;
	}
	else {
		rccb = ctx->rccb[i];
		ctx->rccb[i] = (ritas_rccb_t *)malloc(sizeof(ritas_rccb_t));
		ctx->rccb[i]->next = rccb;
		rccb->prev = rccb;
	}
	
	rccb->prev = NULL;
	rccb->rcid = rcid;
	rccb->lastsn = 0;
	rccb->unack = ritas_ht_init(RITAS_RC_UNACK_SIZE);
	memset(rccb->unack, '\0', RITAS_RC_UNACK_SIZE*sizeof(ritas_rc_unack_t *));
	
	pthread_mutex_init(&rccb->unack_mutex, NULL);
	
	return rccb;
}

void ritas_rc_destroy(ritas_t *ctx, ritas_rccb_t *rccb) {
	
	/*
	 * Unlink the rccb from the list.
	 * My head hurts.
	 */
	if(rccb->prev)
		rccb->prev->next = rccb->next;
	else
		ctx->rccb[RITAS_RC_HASH(rccb->rcid)] = rccb->next;

	if(rccb->next)
		rccb->next->prev = rccb->prev;
	else if(rccb->prev)
		rccb->prev->next = NULL;

	/*
	 * Destroy the unacknowledged packets list.
	 */
	//ritas_mbuf_queue_destroy(rccb->unack);
	
	free(rccb);
	
	return;
}


int ritas_rc_output(ritas_t *ctx, ritas_rccb_t *rccb, ritas_mbuf_t *m) {
	
	struct ritas_rc_header *rc_h;
	ritas_rc_unack_t *pm;
	ritas_proc_t *proc;
	ritas_mbuf_t *mtmp;
	int i, n, hash;
	
	/* build header */
	m->head = m->head - sizeof(struct ritas_rc_header);
	assert(m->head >= m->data);
	m->ttl = ctx->group_size;
	rc_h = (struct ritas_rc_header *)m->head; 
	rc_h->type = RITAS_RC_DATA;
	rc_h->rcid = rccb->rcid;
	rc_h->sn = ++rccb->lastsn;
	
	
	/* 
	 * Rock on. Initialize RTT stuff and send the message to all processes.
	 */
	// TODO: these cycles suck. Must do something about them.
	for(i = 0 ; i < RITAS_GROUP_SIZE ; i++) {
		for(proc = ctx->group[i] ; proc ; proc = proc->next) {
			
			n = ritas_send(ctx, proc, m);
			if(n < 0)
				return -1;
			/*
			 * Add proc-mbuf tuple to unacknowledged message list
			 */
			 //TODO: this probably has to be done before calling ritas_send() in
			 //order to avoid race conditions -> there's a slim chance there may
			 //exist a concurrent ritas_rc_input() processing an 
			 //acknowledgedment for this message that got here super fast
			pm = ritas_rc_unack_init(proc, m);
			hash = RITAS_RC_UNACK_HASH(proc, rc_h->sn);
			pthread_mutex_lock(&rccb->unack_mutex);
			ritas_ht_add(rccb->unack, RITAS_RC_UNACK_SIZE, hash, pm);
			pthread_mutex_unlock(&rccb->unack_mutex);
			
			//ritas_rc_unack_add(rccb->unack, m, proc);		
		}
	}
		
	return 1;
}

void ritas_rc_input(ritas_t *ctx, ritas_mbuf_t *m, ritas_proc_t *proc) {
	
	ritas_rccb_t *rccb;
	ritas_rc_header_t *rc_h, *ack_h;
	ritas_mbuf_t *unack, *m_ack;
	ritas_rc_unack_t *pm;
	int hash;	
	
	rc_h = (ritas_rc_header_t *)m->head;
	
	/*
	 * Locate the control block for this reliable channel.
	 */
	for(rccb = ctx->rccb[RITAS_RC_HASH(rc_h->rcid)] ; rccb ; rccb = rccb->next)
		if(rccb->rcid == rc_h->rcid)
			break;
		
	/*
	 * Hey. We didn't found the control block. Discard this datagram and
	 * get out.
	 */
	// TODO: We may want to hold on to the mbuf for a while even if there's
	// no control block
	if(!rccb) {
		ritas_mbuf_destroy(m);
		return;
	}

	/*
	 * It's an acknowledgement. Let's remove the refered datagram from
	 * the unacknowledged list and get out.
	 */	
	if(rc_h->type == RITAS_RC_ACK) {
		hash = RITAS_RC_UNACK_HASH(proc, rc_h->ack);
		pthread_mutex_lock(&rccb->unack_mutex);
		pm = ritas_ht_remove(rccb->unack, RITAS_RC_UNACK_SIZE, hash);
		pthread_mutex_unlock(&rccb->unack_mutex);
		if(pm) {
			pm->mbuf->ttl--;
			if(pm->mbuf->ttl <= 0)
				ritas_mbuf_destroy(pm->mbuf);
			ritas_rc_unack_destroy(pm);
		}
		ritas_mbuf_destroy(m);
	}
	else {
		/*
		 * We are going to send the acknowledgement.
		 */
		m_ack = ritas_mbuf_init();
	
		m_ack->head = m_ack->tail - sizeof(struct ritas_rc_header);
		assert(m_ack->head >= m_ack->data);
		
		ack_h = (struct ritas_rc_header *)m_ack->head;
		ack_h->type = RITAS_RC_ACK;
		ack_h->rcid = rccb->rcid;
		ack_h->ack = rc_h->sn;
	
		ritas_send(ctx, proc, m_ack);
		ritas_mbuf_destroy(m_ack);
		
		//TODO: check if it's a retransmission of a previously acknowledged datagram
		ritas_rb_input(ctx, rccb, m, proc);
		ritas_mbuf_destroy(m); // for now destroy the mbuf
	}	

}
