#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <pthread.h>
#include <openssl/md5.h>
#include "ritas_ab.h"

int ritas_ab_input(ritas_t *ctx, ritas_abcb_t *cb, ritas_mbuf_t *m) {

	ritas_ab_header_t abh, *abhp;
	ritas_ab_aht_entry_t *e;
	ritas_rbcb_t *rbcb;
	u_short txid;
	int n, i, j;
	ritas_vector_t *X, *Y;
	ritas_mbuf_t *m_rdelivered;
	ritas_ab_vproc_t *vproc;
	
	/*
	 * If the received message is a new Broadcast, then we may have one of two
	 * possible scenarios:
	 * (a) a VC instance isn't running. This means we add the message to the
	 *     R_Delivered table and start a new VC instance proposing the messages
	 *     in the R_Delivered table.
	 * 
	 * (b) a VC instance is already running. Check if the received message is
	 *     in f+1 cells of the decided vector (if case we already have one), if
	 *     so add it to the A_Deliver table. Otherwise, just add the message to
	 *     the R_Delivered table.
	 */
	if(m->protocol == RITAS_MBUF_RB) {
	
		rbcb = m->rbcb;
		txid = m->proc->id;
		/*
		 * Calculate hash of received message.
		 */
		//MD5(m->head, m->tail - m->head, m->hash);
		//hash = m->hash;
		//hash_s = MD5_DIGEST_LENGTH;
		abh.txid = m->id->txid;
		abh.num = m->id->rbid;

		dprintf("[%d] AB in ID(%d,%d)\n", ctx->pid, abh.txid, abh.num);

		/*
		 * Renew RBCB.
		 */
		n = ritas_rb_renew(ctx, rbcb, txid, ++cb->rb_num[txid]);
		if(n == -1) {
			snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): unable to renew cb\n", __func__);
			return -1;
		}
		
		if(txid == ctx->pid) {
			pthread_mutex_lock(&cb->rbcb_mutex);
			ritas_list_enqueue(cb->rbcb_l, rbcb);
			cb->rbcb_count++;
			pthread_cond_broadcast(&cb->rbcb_cond);
			pthread_mutex_unlock(&cb->rbcb_mutex);
		}		

		if(cb->lock == 0) {
			cb->lock = 1;
			/*
			 * Insert message in R_Delivered table.
			 */
			n = ritas_ab_aht_insert(cb->r_aht, (u_char *)&abh, sizeof(abh), m);
			if(n == -1) {
				snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): unable to add element to augmented hash table\n", __func__);
				return -1;
			}
			cb->r_count++;

			n = ritas_ab_consensus(ctx, cb);
			if(n == -1)
				return -1;	
		}
		else if(cb->lock == 1){
			/*
			 * Lookup this hash in the A_deliver table.
			 */	
			e = ritas_ab_aht_lookup(cb->a_aht, (u_char *)&abh, sizeof(abh));
			if(e) {
				/*
				 * This message has its hash is in the A_deliver table. It means we were
				 * waiting for this message to be atomically delivered. Pass it right
				 * away to the A_deliver table.
				 */
				ritas_ab_aht_update(e, m);
				cb->a_count--;
			}
			else {
				/*
				 * Insert message in R_Delivered table.
			 	*/
				n = ritas_ab_aht_insert(cb->r_aht, (u_char *)&abh, sizeof(abh), m);
				if(n == -1) {
					snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): unable to add element to augmented hash table\n", __func__);
					return -1;
				}
				cb->r_count++;
			}
		}
	}
	/*
	 * If the received message is a VC decision, then count the hashes and build
	 * a table with the hashes (and messages in case they're already in
	 * R_Delivered) that appear in f+1 or more cells.
	 */
	else if(m->protocol == RITAS_MBUF_VC) {
		cb->decision = 1;
		/*
		 * Iterate over the agreed vector.
		 */
		X = ritas_vector_extract(m->head, m->tail - m->head);
		if(!X)
			return -1;

		dprintf("[%d] AB agreed to deliver: ", ctx->pid);
		for(i = 0 ; i < X->size ; i++) {
			if(X->element_s[i] > 0) {
				Y = ritas_vector_extract(X->element[i], X->element_s[i]);
				if(!Y) {
					snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): unable to extract vector\n", __func__);
					return -1;
				}
				
				for(j = 0 ; j < Y->size ; j++) {
					if(Y->element_s[j] > 0) {
						/*
						 * Count the number of times this given hash appeared.
						 */
						vproc = ritas_ht_lookup(cb->v_count, Y->element[j], Y->element_s[j]);
						if(!vproc) {
							vproc = ritas_ab_vproc_init(ctx->n);
							if(!vproc) {
								snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): unable to init vproc\n", __func__);
								return -1;
							}
						
							n = ritas_ht_add(cb->v_count, Y->element[j], Y->element_s[j], vproc);
							if(n == -1) {
								snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): unable to add element to hash table\n", __func__);
								return -1;
							}
							
						}	
						if(!vproc->proc[i]) {
							vproc->count++;
							vproc->proc[i] = 1;
						}
						abhp = (ritas_ab_header_t *)Y->element[j];
						dprintf("(%d,%d)[%d]", abhp->txid, abhp->num, vproc->count);
						/*
						 * If the count for this given hash reaches f+1, then we want
						 * to atomically deliver the corresponding message.
						 */
						if(vproc->count == ctx->f+1) {
							dprintf("*");
							/*
							 * Transfer the message from the R_delivered table (if it's
							 * actually there) and update the message counts.
							 */
							m_rdelivered = ritas_ab_aht_remove(cb->r_aht, Y->element[j], Y->element_s[j]);
							if(m_rdelivered)
								cb->r_count--;

							n = ritas_ab_aht_insert(cb->a_aht, Y->element[j], Y->element_s[j], m_rdelivered);
							if(n == -1) {
								snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): unable to insert element in augmented hash table\n", __func__);
								return -1;
							}
							if(!m_rdelivered)
								cb->a_count++;
						}
						dprintf(" ");
					}
				}
				ritas_vector_destroy(Y);
			}
		}
		dprintf("\n");
		ritas_vector_destroy(X);
		ritas_mbuf_destroy(m);
		
		ritas_ht_renew(cb->v_count, (void(*)(void *))ritas_ab_vproc_destroy);
	}
	
	/*
	 * If the count in the A_deliver table reaches zero, it means we have all
	 * the messages we agreed to atomically deliver in the last VC instante.
	 * Then just do it, stupid!
	 */
	if(cb->decision == 1 && cb->a_count == 0) {
		dprintf("[%d] AB(%d) all the messages we agreed on VC are here\n", ctx->pid, cb->vcid);
	
		/*
		 * Put the messages in the atomic delivery queue.
		 *
		 */
		pthread_mutex_lock(&cb->a_delivered_mutex);
		ritas_list_concat(cb->a_delivered, ritas_ab_aht_list(cb->a_aht));
		pthread_cond_broadcast(&cb->a_delivered_cond);
		pthread_mutex_unlock(&cb->a_delivered_mutex);
		ritas_ab_aht_renew(cb->a_aht);

		dprintf("[%d] AB(%d) consensus #%d completed\n", ctx->pid, cb->vcid, cb->vcid);
		
		cb->vcid++;
		cb->lock = 0;
		cb->decision = 0;
		
		/*
		 * Start a new vector consensus.
		 */
		ritas_vc_destroy(ctx, cb->vccb);
		cb->vccb = ritas_vc_init(ctx, cb->vcid, cb);
		if(!cb->vccb)
			return -1;
			
		if(cb->r_count > 0) {
			cb->lock = 1;
			n = ritas_ab_consensus(ctx, cb);
			if(n == -1)
				return -1;
		}
	}		
	
	return 1;
}


int ritas_ab_output(ritas_t *ctx, ritas_abcb_t *cb, ritas_mbuf_t *m) {

	ritas_rbcb_t *rbcb;
	ritas_ab_header_t *abh;
	int n;

	/*
	 * Wait for a free RBCB.
	 */
	pthread_mutex_unlock(&ctx->temp_mutex);	
	pthread_mutex_lock(&cb->rbcb_mutex);
	while(!cb->rbcb_count)
		pthread_cond_wait(&cb->rbcb_cond, &cb->rbcb_mutex);
	rbcb = ritas_list_dequeue(cb->rbcb_l);
	cb->rbcb_count--;
	pthread_mutex_unlock(&cb->rbcb_mutex);
	pthread_mutex_lock(&ctx->temp_mutex);
		
	/*
	 * Build AB header.
	 */
	m->head = m->head - sizeof(*abh);
	if(m->head < m->data) {
		snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): mbuf data overflow\n", __func__);
		return -1;
	}
	abh = (ritas_ab_header_t *)m->head;
	/*
	 * TODO: correct this
	 */
	abh->num = rbcb->rbid;
	abh->txid = rbcb->txid;
	
	/*
	 * And put the wheels in motion.
	 */
	n = ritas_rb_output(ctx, rbcb, m);
	if(n == -1)
		return -1;
		
	return 1;

}

int ritas_ab_consensus(ritas_t *ctx, ritas_abcb_t *cb) {

	ritas_vector_t *H;
	ritas_list_t *l;
	ritas_mbuf_t *m, *m_out;
	int n;
	
	/*
	 * Init vector H.
	 */
	H = ritas_vector_init(cb->r_count);
	if(!H) {
		snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): unable to init vector\n", __func__);
		return -1;
	}
	
	/*
	 * Build message H with hashes of messages in R_delivered.
	 */
	m_out = ritas_mbuf_init();
	if(!m_out) {
		snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): unable to init mbuf\n", __func__);
		return -1;
	}
	
	l = ritas_ab_aht_list(cb->r_aht);
	while((m = ritas_list_dequeue(l))) {
		n = ritas_vector_add(H, (u_char *)&m->id->txid, sizeof(ritas_ab_header_t));
		if(n == -1) {
			snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): unable to add element to vector\n", __func__);
			return -1;
		}				
	}
	dprintf("\n");
	ritas_list_destroy(l);
	
	m_out->head = m_out->head - H->data_s;
	if(m_out->head < m_out->data) {
		snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): mbuf data overflow\n", __func__);
		return -1;
	}
			
	ritas_vector_build(H, m_out->head);
	ritas_vector_destroy(H);
	n = ritas_vc_output(ctx, cb->vccb, m_out);
	if(n == -1)
		return -1;

	return 1;
}


int ritas_ab_init(ritas_t *ctx) {

	ritas_abcb_t *cb;
	int i, j, x;

	pthread_mutex_lock(&ctx->temp_mutex);
	
	/*
	 * Allocate and zero out CB.
	 */
	cb = malloc(sizeof(*cb));
	if(!cb) {
		snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): unable to malloc() control block\n", __func__);
		goto bad;
	}
	memset(cb, '\0', sizeof(*cb));
	ctx->abcb = cb;

	/*
	 * Set ID to 1.
	 */
	cb->id = 1;
	cb->vcid = 1;
	cb->lock = 0;		
	cb->decision = 0;
		
	/*
	 * Init atomically delivered message list, mutex and condition variable.
	 */
	cb->a_delivered = ritas_list_init();
	if(!cb->a_delivered) {
		snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): unable to init list\n", __func__);
		goto bad;
	}
	pthread_mutex_init(&cb->a_delivered_mutex, NULL);
	pthread_cond_init(&cb->a_delivered_cond, NULL);

	/*
	 * Init consensus message augmented hash table, count mutex and condition
	 * variable.
	 */	
	cb->a_aht = ritas_ab_aht_init();
	if(!cb->a_aht) {
		snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): unable to init augmented hash table\n", __func__);
		goto bad;
	}
	cb->a_count = 0;
	
	cb->r_aht = ritas_ab_aht_init();
	cb->r_count = 0;
	
	/*
	 * Init value count hash table.
	 */
	cb->v_count = ritas_ht_init(RITAS_AB_V_COUNT_SIZE);
	if(!cb->v_count) {
		snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): unable to init hash table\n", __func__);
		goto bad;
	}

	/*
	 * Init VC control block.
	 */
	cb->vccb = ritas_vc_init(ctx, cb->vcid, cb);
	if(!cb->vccb)
		goto bad;
	
	/*
	 * Alloc RBCB array with size ctx->n*RITAS_AB_THROUGHPUT.
	 */
	cb->rbcb = malloc(ctx->n*RITAS_AB_THROUGHPUT*sizeof(ritas_rbcb_t *));
	if(!cb->rbcb) {
		snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): unable to init array of control blocks\n", __func__);
		goto bad;
	}	

	/*
	 * Init the available RBCBs list.
	 */
	cb->rbcb_l = ritas_list_init();
	if(!cb->rbcb_l) {
		snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): unable to init list\n", __func__);
		goto bad;
	}
		
	/*
	 * Init and enqueue to the available list every RBCB in the array; set the
	 * count to RITAS_AB_THROUGHPUT.
	 */
	for(i = 0; i < ctx->n ; i++) {
		for(j = 0 ; j < RITAS_AB_THROUGHPUT ; j++) {
			x = i*RITAS_AB_THROUGHPUT+j;
			cb->rbcb[x] = ritas_rb_init(ctx, i, j+1, cb, RITAS_RB_UPPER_AB, 0);
			if(!cb->rbcb[x])
				goto bad;
			if(ctx->pid == i)
				ritas_list_enqueue(cb->rbcb_l, cb->rbcb[x]);
		}
	}
	cb->rbcb_count = RITAS_AB_THROUGHPUT;
	
	/*
	 * Init the RB num array with size n; zero out the array.
	 */
	cb->rb_num = malloc(ctx->n*sizeof(*cb->rb_num));
	if(!cb->rb_num) {
		snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): unable to init array\n", __func__);
		goto bad;
	}
	for(i = 0 ; i < ctx->n ; i++) {
		cb->rb_num[i] = RITAS_AB_THROUGHPUT;
	}

	pthread_mutex_unlock(&ctx->temp_mutex);

	/*
	 * Yes!
	 */
	return 1;
	
bad:
	/*
	 * Nooooooo!
	 */
	if(cb) {
		ritas_ab_destroy(ctx);
	}
	return -1;
}

void ritas_ab_destroy(ritas_t *ctx) {
	
	int i, j;
	ritas_abcb_t *cb;

	pthread_mutex_lock(&ctx->temp_mutex);
	
	cb = ctx->abcb;
	if(!cb)
		return;
	
	ritas_vc_destroy(ctx, cb->vccb);

	for(i = 0 ; i < ctx->n ; i++)
		for(j = 0 ; j < RITAS_AB_THROUGHPUT ; j++)
			ritas_rb_destroy(ctx, cb->rbcb[i*RITAS_AB_THROUGHPUT+j]);
	free(cb->rbcb);
	free(cb->rb_num);
	ritas_list_destroy(cb->rbcb_l);
	pthread_mutex_destroy(&cb->rbcb_mutex);
	pthread_cond_destroy(&cb->rbcb_cond);	

	ritas_list_destroy(cb->a_delivered);
	pthread_mutex_destroy(&cb->a_delivered_mutex);
	pthread_cond_destroy(&cb->a_delivered_cond);

	ritas_ab_aht_destroy(cb->a_aht);

	ritas_ab_aht_destroy(cb->r_aht);
	
	ritas_ht_destroy(cb->v_count, (void(*)(void *))ritas_ab_vproc_destroy);
	
	free(cb);
	
	pthread_mutex_unlock(&ctx->temp_mutex);	
}

ritas_ab_vproc_t *ritas_ab_vproc_init(u_short n) {
	
	ritas_ab_vproc_t *vproc;
	
	vproc = malloc(sizeof(*vproc));
	if(!vproc)
		return NULL;

	vproc->proc = malloc(n*sizeof(vproc->proc));
	
	if(!vproc->proc) {
		free(vproc);
		return NULL;
	}
	memset(vproc->proc, '\0', n*sizeof(vproc->proc));
	
	vproc->count = 0;
	
	return vproc;
}

void ritas_ab_vproc_destroy(ritas_ab_vproc_t *vproc) {
	free(vproc->proc);
	free(vproc);
}


int ritas_ab_bcast(ritas_t *ctx, u_char *buf, u_short buf_s) {
	
	ritas_mbuf_t *m;
	int n;
	
	m = ritas_mbuf_init();
	if(!m) {
		snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): unable to init mbuf\n", __func__);
		goto bad;
	}
	
	m->head = m->head - buf_s;
	if(m->head < m->data) {
		snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): mbuf data overflow\n", __func__);
		goto bad;
	}
	memcpy(m->head, buf, buf_s);

	pthread_mutex_lock(&ctx->temp_mutex);
	
	n = ritas_ab_output(ctx, ctx->abcb, m);
	if(n == -1)
		return -1;

	pthread_mutex_unlock(&ctx->temp_mutex);
	
	return buf_s;
	
bad:
	if(m)
		ritas_mbuf_destroy(m);

	return -1;
}

int ritas_ab_recv(ritas_t *ctx, u_char *buf, u_short buf_s) {
	
	ritas_mbuf_t *m;
	u_short len;
	
	pthread_mutex_lock(&ctx->err_mutex);
	if(ctx->error) {
		ctx->error = 0;
		return -1;
	}
	pthread_mutex_unlock(&ctx->err_mutex);
	pthread_mutex_lock(&ctx->abcb->a_delivered_mutex);
	while(!ctx->abcb->a_delivered->first)
		pthread_cond_wait(&ctx->abcb->a_delivered_cond, &ctx->abcb->a_delivered_mutex);
	m = ritas_list_dequeue(ctx->abcb->a_delivered);
	pthread_mutex_unlock(&ctx->abcb->a_delivered_mutex);

	len = m->tail - m->head;
	if(len > buf_s) {
		snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): buffer overflow\n", __func__);
		return -1;
	}
	memcpy(buf, m->head, len);
	
	ritas_mbuf_destroy(m);
	
	return len;
}

u_short ritas_ab_id(ritas_abcb_t *cb) {
	return cb->id;
}
