#include <stdlib.h>
#include <string.h>
#include "ritas_ht.h"

ritas_ht_t *ritas_ht_init(u_char tsize) {
	
	ritas_ht_t *ht;
	
	ht = malloc(sizeof(ritas_ht_t));
	if(!ht)
		return NULL;
	
	ht->table = malloc(tsize*sizeof(ritas_ht_entry_t *));
	if(!ht->table) {
		free(ht);
		return NULL;
	}
		
	memset(ht->table, '\0', tsize*sizeof(ritas_ht_entry_t *));
	ht->size = tsize;
	
	return ht;
}

void ritas_ht_destroy(ritas_ht_t *ht, void (*destroy_func)(void *)) {
	
	if(!ht)
		return;
		
	ritas_ht_renew(ht, destroy_func);

	free(ht->table);
	free(ht);
}

void ritas_ht_renew(ritas_ht_t *ht, void (*destroy_func)(void *v)) {

	int i;
	ritas_ht_entry_t *httmp;
	
	if(!ht)
		return;
		
	for(i = 0 ; i < ht->size ; i++) {
		while(ht->table[i]) {
			httmp = ht->table[i];
			ht->table[i] = ht->table[i]->next;
			
			if(destroy_func)
				(*destroy_func)(httmp->value);
			else
				free(httmp->value);

			free(httmp->key);				
			free(httmp);
		}
	}
}

int ritas_ht_add(ritas_ht_t *ht, u_char *key, u_short ksize, void *value) {

	ritas_ht_entry_t *httmp;
	u_char index;
	
	index = ritas_ht_hash(ht, key, ksize);
	
	httmp = malloc(sizeof(ritas_ht_entry_t));
	if(!httmp)
		return -1;
	
	httmp->value = value;
	httmp->ksize = ksize;
	httmp->key = malloc(httmp->ksize);
	if(!httmp->key) {
		free(httmp);
		return -1;
	}
	memcpy(httmp->key, key, ksize);

	httmp->next = ht->table[index];
	ht->table[index] = httmp;

	return 1;
	

}

int ritas_ht_add_ext(ritas_ht_t *ht, u_char *key, u_short ksize, void *value) {

	ritas_ht_entry_t *httmp;
	u_char index;
	
	index = ritas_ht_hash(ht, key, ksize);
	
	httmp = malloc(sizeof(ritas_ht_entry_t));
	if(!httmp)
		return -1;
	
	httmp->value = value;
	httmp->ksize = ksize;
	httmp->key = key;

	httmp->next = ht->table[index];
	ht->table[index] = httmp;

	return 1;
}	
	

void *ritas_ht_remove(ritas_ht_t *ht, u_char *key, u_short ksize) {

	ritas_ht_entry_t *httmp;
	void *ret;
	
	httmp = ritas_ht_remove_ext(ht, key, ksize);
	if(!httmp)
		return NULL;
	
	ret = httmp->value;
	free(httmp);
	return ret;
}

ritas_ht_entry_t *ritas_ht_remove_ext(ritas_ht_t *ht, u_char *key, u_short ksize) {

	u_char index;
	ritas_ht_entry_t *httmp, *htprev;
	
	index = ritas_ht_hash(ht, key, ksize);
	
	htprev = httmp = NULL;
			
	for(httmp = ht->table[index] ; httmp ; htprev = httmp, httmp = httmp->next) {
		if(ksize == httmp->ksize) {
			if(!memcmp(httmp->key, key, ksize)) {
				if(!htprev)
					ht->table[index] = httmp->next;
				else
					htprev->next = httmp->next;
				free(httmp->key);
				return httmp;
			}
		}
	}
	
					
	return NULL;
}

void *ritas_ht_lookup(ritas_ht_t *ht, u_char *key, u_short ksize) {

	u_char index;
	ritas_ht_entry_t *httmp;
	
	httmp = ritas_ht_lookup_ext(ht, key, ksize);
	if(!httmp)
		return NULL;
	
	return httmp->value;
}

ritas_ht_entry_t *ritas_ht_lookup_ext(ritas_ht_t *ht, u_char *key, u_short ksize) {

	u_char index;
	ritas_ht_entry_t *httmp;
	
	index = ritas_ht_hash(ht, key, ksize);
			
	for(httmp = ht->table[index] ; httmp ; httmp = httmp->next)
		if(ksize == httmp->ksize)
			if(!memcmp(httmp->key, key, ksize))
				return httmp;
	
	return NULL;
}

u_char ritas_ht_hash(ritas_ht_t *ht, u_char *key, u_short ksize) {
	int i;
	u_char index;

	for(index = 0, i = 0 ; i < ksize ; i++)
		index += key[i];
	index = index % ht->size;

	return index;
}
