#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <netinet/tcp.h>
#include <pthread.h>
#include "ritas.h"
#include "ritas_ab.h"
#include "ritas_rb.h"
#include "ritas_rc.h"
#include "ritas_proc.h"
#include "ritas_mbuf.h"

ritas_t *ritas_init(u_short pid, u_short n, u_short f, u_short port, char *errbuf) {
	
	ritas_t *ctx;
	struct sockaddr_in addr;
	int r;
#ifdef TCP
	struct linger lg;
	int sockbuf;
#endif
	
	ctx = (ritas_t *)malloc(sizeof(ritas_t));
	if(!ctx) {
        snprintf(errbuf, RITAS_ERRBUF_SIZE, "%s(): unable to malloc() context\n", __func__);
		return NULL;
	}

	memset(ctx, '\0', sizeof(*ctx));
	
#ifdef TCP
	ctx->socket = socket(AF_INET, SOCK_STREAM, 0);
	lg.l_onoff = 1;
	lg.l_linger = 0;
	setsockopt(ctx->socket, SOL_SOCKET, SO_LINGER, (char *)&lg, sizeof(lg));
	sockbuf = 1024*1024;
	setsockopt(ctx->socket, SOL_SOCKET, SO_SNDBUF, &sockbuf, sizeof(sockbuf));
	setsockopt(ctx->socket, SOL_SOCKET, SO_RCVBUF, &sockbuf, sizeof(sockbuf));
#else	
	ctx->socket = socket(AF_INET, SOCK_DGRAM, 0);
#endif	
	if(ctx->socket < 0) {
        snprintf(errbuf, RITAS_ERRBUF_SIZE, "%s(): socket()\n", __func__);
		return NULL;
	}
	
	memset(&addr, '\0', sizeof(addr));
	addr.sin_family = AF_INET;
	addr.sin_port = htons(port);
	addr.sin_addr.s_addr = htonl(INADDR_ANY);
	
	r = bind(ctx->socket, (struct sockaddr *)&addr, sizeof(addr));
	if(r < 0) {
        snprintf(errbuf, RITAS_ERRBUF_SIZE, "%s(): bind()\n", __func__);
		ritas_destroy(ctx);
		return NULL;
	}

#ifdef TCP
	r = listen(ctx->socket, n);
	if(r < 0) {
        snprintf(errbuf, RITAS_ERRBUF_SIZE, "%s(): listen()\n", __func__);
		return NULL;
	}
#endif	
	
	ctx->ht_rccb = ritas_ht_init(RITAS_HT_RCCB_SIZE);
	if(!ctx->ht_rccb) {
		snprintf(errbuf, RITAS_ERRBUF_SIZE, "%s(): unable to initialize hash table\n", __func__);
		ritas_destroy(ctx);
		return NULL;
	}

	ctx->ht_ooc = ritas_ht_init(RITAS_HT_OOC_SIZE);
	if(!ctx->ht_ooc) {
		snprintf(errbuf, RITAS_ERRBUF_SIZE, "%s(): unable to initialize hash table\n", __func__);
		ritas_destroy(ctx);
		return NULL;
	}
	
	ctx->l_ooc = ritas_list_init();
	if(!ctx->l_ooc) {
		snprintf(errbuf, RITAS_ERRBUF_SIZE, "%s(): unable to initialize linked list\n", __func__);		
		ritas_destroy(ctx);
		return NULL;
	}
	
	ctx->n = n;
	ctx->f = f;
	ctx->pid = pid;
	ctx->group = (ritas_proc_t **)malloc(n*sizeof(ritas_proc_t *));
	if(!ctx->group) {
        snprintf(errbuf, RITAS_ERRBUF_SIZE, "%s(): unable to malloc() group array\n", __func__);		
		ritas_destroy(ctx);
		return NULL;
	}
	memset(ctx->group, '\0', n*sizeof(ritas_proc_t *));

	ctx->error = 0;
	pthread_mutex_init(&ctx->err_mutex, NULL);
	ctx->aproxy = NULL;

	ctx->scheduler_l = ritas_list_init();
	if(!ctx->scheduler_l) {
        snprintf(errbuf, RITAS_ERRBUF_SIZE, "%s(): ritas_list_init()\n", __func__);
		ritas_destroy(ctx);
		return NULL;
	}		

	ctx->scheduler_v = 0;
	pthread_mutex_init(&ctx->scheduler_mutex, NULL);
	pthread_cond_init(&ctx->scheduler_cond, NULL);

	pthread_mutex_init(&ctx->temp_mutex, NULL);

	ctx->pr_count = 0;
	pthread_mutex_init(&ctx->pr_mutex, NULL);
	pthread_cond_init(&ctx->pr_cond, NULL);
		
	gettimeofday(&ctx->t_init, NULL);	
		
	return ctx;
}

void ritas_destroy(ritas_t *ctx) {

#ifdef TCP	
	int i;
#endif
	
#ifndef TCP
	close(ctx->socket);
#endif

	for(i = 0 ; i < ctx->n ; i++) {
#ifdef TCP
		close(ctx->group[i]->socket);
#endif
		free(ctx->group[i]);
	}
	free(ctx->group);
	
	ritas_ht_destroy(ctx->ht_ooc, (void(*)(void *))&ritas_list_destroy_mbuf);
	ritas_ht_destroy(ctx->ht_rccb, (void(*)(void *))&ritas_rb_destroy);
	ritas_list_destroy(ctx->l_ooc);
	
	pthread_mutex_destroy(&ctx->err_mutex);
	
	ritas_list_destroy(ctx->scheduler_l);
	
	pthread_mutex_destroy(&ctx->scheduler_mutex);
	pthread_cond_destroy(&ctx->scheduler_cond);

	pthread_mutex_destroy(&ctx->pr_mutex);
	pthread_cond_destroy(&ctx->pr_cond);
	
	free(ctx);
}

void ritas_renew(ritas_t *ctx) {

	pthread_mutex_lock(&ctx->temp_mutex);

	ritas_ht_renew(ctx->ht_ooc, (void(*)(void *))&ritas_list_destroy_mbuf);
	ritas_ht_renew(ctx->ht_rccb, (void(*)(void *))&ritas_rb_destroy);
	
	pthread_mutex_unlock(&ctx->temp_mutex);
}


int ritas_start(ritas_t *ctx) {

	int n;
#ifdef TCP
	int i, s;
	struct sockaddr_in peeraddr;
	socklen_t addrlen;
	u_short id;
	int flag;
	
	for(i = 0 ; i < ctx->pid ; i++) {
		for(;;) {
			n = connect(ctx->group[i]->socket, (struct sockaddr *)&ctx->group[i]->addr, sizeof(ctx->group[i]->addr));
			if(n < 0) {
    		    if(errno != ECONNREFUSED) {
	    		    snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): connect()\n", __func__);
					return -1;
    		    }
			}
			else
				break;
		}
		
		n = send(ctx->group[i]->socket, &ctx->pid, sizeof(ctx->pid), 0);
		if(n < 0) {
    	    snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): send()\n", __func__);
			return -1;
		}
	}

	addrlen = sizeof(peeraddr);
	for(i = ctx->pid+1 ; i < ctx->n ; i++) {
		s = accept(ctx->socket, (struct sockaddr *)&peeraddr, &addrlen);
		if(s < 0) {
    	    snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): accept()\n", __func__);
			return -1;
		}
		
		flag = 1;
		setsockopt(s, IPPROTO_TCP, TCP_NODELAY, (char *)&flag, sizeof(flag));		
		
		n = recv(s, &id, sizeof(id), 0);
		if(n < 0) {
    	    snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): recv()\n", __func__);
			return -1;
		}
		dprintf("accepted connection from %d\n", id);
		// TODO: cryptographically strong integrity check or enforce port determinism
		ctx->group[id]->socket = s;
	}
	close(ctx->socket);
	
#endif
/*	
	for(i = 0 ; i < ctx->n ; i++) {
		n = pthread_create(&ctx->net_thread[i], NULL, (void *)&ritas_net, (void *)ctx);
		if(n) {
	        snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): pthread_create()\n", __func__);
			return -1;
		}
	}	
*/	
	n = pthread_create(&ctx->scheduler_thread, NULL, (void *)&ritas_scheduler, (void *)ctx);
	if(n) {
        snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): pthread_create()\n", __func__);
		return -1;
	}


	n = pthread_create(&ctx->scheduler_net_thread, NULL, (void *)&ritas_scheduler_net, (void *)ctx);
	if(n) {
        snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): pthread_create()\n", __func__);
		return -1;
	}

	
	return 1;
}


void ritas_scheduler_net(ritas_t *ctx) {
	
	struct timeval t0, t1;	
	ritas_mbuf_t *m;
	int n;
#ifndef TCP
	struct sockaddr_in addr;
	socklen_t addrlen;
#endif
#ifdef TCP
	int i, d, b;
	fd_set rset, nset;
	u_short len;
	int bytes;
	int maxsockp1;
#endif	
	
#ifdef TCP
	FD_ZERO(&nset);
	maxsockp1 = 0;
	for(i = 0 ; i < ctx->n ; i++) {
		if(i == ctx->pid)
			continue;
		FD_SET(ctx->group[i]->socket, &nset);
		if(ctx->group[i]->socket > maxsockp1)
			maxsockp1 = ctx->group[i]->socket;
	}
	maxsockp1++;
#else
	addrlen = sizeof(addr);
#endif

	for(;;) {
#ifdef TCP
		rset = nset;

		d = select(maxsockp1, &rset, NULL, NULL, NULL);
		if(d < 0) {
			printf("ULAHLAH\n");
			snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): select()\n", __func__);
			break;
		}

		
		for(i = 0 ; i < ctx->n ; i++) {
			if(FD_ISSET(ctx->group[i]->socket, &rset)) {				
				
				ioctl(ctx->group[i]->socket, FIONREAD, &bytes);
				if(!bytes) {
					FD_CLR(ctx->group[i]->socket, &nset);
				}
				if(bytes < sizeof(len))
					continue;
				
				/*
				 * Check the packet size
				 */
				n = recv(ctx->group[i]->socket, &len, sizeof(len), MSG_PEEK);
				if(bytes-sizeof(len) < len)
					continue;
				
				m = ritas_mbuf_init();
				if(!m)
					break;				
				
				/*
				 * Read the packet then
				 */
				ioctl(ctx->group[i]->socket, FIONREAD, &bytes);				 
				n = recv(ctx->group[i]->socket, &len, sizeof(len), 0);
				n = recv(ctx->group[i]->socket, m->data, len, 0);
				
/*				
				n = recv(ctx->group[i]->socket, &len, sizeof(len), MSG_PEEK);
				if(n < 0) {
					dprintf("ALERT\n");
					snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): recv()\n", __func__);
					break;
				}
				else if(n == 0) {
					dprintf("n == 0 -> CLEARING process %d\n", i);
					FD_CLR(ctx->group[i]->socket, &nset);
					continue;
				}
				if(n < sizeof(len))
					continue;


				n = recv(ctx->group[i]->socket, &len, sizeof(len), 0);
				if(n != sizeof(len)) {
					printf("ALERT recv(1) n(%d) != len(%d)\n", n, sizeof(len));
					exit(1);
				}

				m = ritas_mbuf_init();
				if(!m)
					break;
				
				n = recv(ctx->group[i]->socket, m->data, len, MSG_WAITALL);		
				if(n < 0) {
					snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): recv()\n", __func__);
					printf("%s(): recv()\n", __func__);
					exit(1);
				}
					
				if(n != len) {
					printf("ALERT recv(2) n(%d) != len(%d)\n", n, len);
					if(n == 0) {
						printf("ALERT n==0 UHOH\n");
						FD_CLR(ctx->group[i]->socket, &nset);
					}
					exit(1);
				}
*/				
				
				m->head = m->data;
				m->tail = m->data + n;		
				
				pthread_mutex_lock(&ctx->scheduler_mutex);
				ritas_list_enqueue(ctx->scheduler_l, m);
				ctx->scheduler_v = 1;
				pthread_cond_broadcast(&ctx->scheduler_cond);
				pthread_mutex_unlock(&ctx->scheduler_mutex);

				m = NULL;

				d--;
				if(d == 0)
					break;
			}

		}
#else	
		if(m) {	
			m = ritas_mbuf_init();
			if(!m)
				break;
		}
				
		n = recvfrom(ctx->socket, m->data, RITAS_MBUF_DATA_SIZE, 0, (struct sockaddr *)&addr, &addrlen);
		if(n < 0) {
		    if(errno == ECONNRESET)
		    	continue;
		    snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): recv()\n", __func__);		    	
			break;
		}
		
		m->head = m->data;
		m->tail = m->data + n;
		
		pthread_mutex_lock(&ctx->scheduler_mutex);
		ritas_list_enqueue(ctx->scheduler_l, m);
		ctx->scheduler_v = 1;
		pthread_cond_broadcast(&ctx->scheduler_cond);
		pthread_mutex_unlock(&ctx->scheduler_mutex);
#endif
	}
	
	/*
	 * Big beef. Notify application via aproxy.
	 */
	ritas_mbuf_destroy(m);
	pthread_mutex_lock(&ctx->err_mutex);
	ctx->error = 1;
	dprintf("%s\n", ctx->errbuf);
	pthread_mutex_unlock(&ctx->err_mutex);
	/*
	ctx->aproxy->mvalue = NULL;
	pthread_mutex_lock(&ctx->aproxy->mutex);
	ctx->aproxy->count++;
	pthread_cond_broadcast(&ctx->aproxy->cond);
	pthread_mutex_unlock(&ctx->aproxy->mutex);
	*/
}

void ritas_scheduler(ritas_t *ctx) {

	ritas_list_t *mlist;
	ritas_mbuf_t *m;
	ritas_rccb_t *rccb;
	ritas_id_t id;
	int n;
	struct sockaddr_in addr;
	socklen_t addrlen;

	for(;;) {
		/*
		 * Check for OOC messages.
		 */
		pthread_mutex_lock(&ctx->temp_mutex);	
		while((rccb = ritas_list_dequeue(ctx->l_ooc))) {
			id = rccb->id;
			if((mlist = ritas_ht_lookup(ctx->ht_ooc, (u_char *)&id, RITAS_ID_HEADER_SIZE))) {
				while((m = ritas_list_dequeue(mlist))) {
					if(rccb->parent_proto == RITAS_RC_PROTO_EB)
						n = ritas_eb_input(ctx, rccb->parent, m);
					else if(rccb->parent_proto == RITAS_RC_PROTO_RB)
						n = ritas_rb_input(ctx, rccb->parent, m);
					else if(rccb->parent_proto == RITAS_RC_PROTO_BC)
						n = ritas_bc_input(ctx, rccb->parent, m);
					if(n == -1)
						break;
				}
			}
			ritas_ht_remove(ctx->ht_ooc, (u_char *)&id, RITAS_ID_HEADER_SIZE);
			ritas_list_destroy(mlist);
		}			
		pthread_mutex_unlock(&ctx->temp_mutex);
			
		pthread_mutex_lock(&ctx->scheduler_mutex);
		while(!ctx->scheduler_v)
			pthread_cond_wait(&ctx->scheduler_cond, &ctx->scheduler_mutex);
		m = ritas_list_dequeue(ctx->scheduler_l);
		if(!ctx->scheduler_l->first)
			ctx->scheduler_v = 0;
		pthread_mutex_unlock(&ctx->scheduler_mutex);
		
		if(m) {
			pthread_mutex_lock(&ctx->temp_mutex);			
			n = ritas_rc_input(ctx, m);
			pthread_mutex_unlock(&ctx->temp_mutex);
			if(n == -1)
				break;
		}

		
	}
	
	/*
	 * Big beef. Notify application via aproxy.
	 */
	ritas_mbuf_destroy(m);
	pthread_mutex_lock(&ctx->err_mutex);
	ctx->error = 1;
	dprintf("%s\n", ctx->errbuf);
	pthread_mutex_unlock(&ctx->err_mutex);
	/*
	ctx->aproxy->mvalue = NULL;
	pthread_mutex_lock(&ctx->aproxy->mutex);
	ctx->aproxy->count++;
	pthread_cond_broadcast(&ctx->aproxy->cond);
	pthread_mutex_unlock(&ctx->aproxy->mutex);
	*/
}

void ritas_timestamp() {
	struct timeval t;
	
	gettimeofday(&t, NULL);
	printf("%ld.%ld ", t.tv_sec, t.tv_usec);
}

#ifdef TCP
void ritas_accept(ritas_t *ctx) {
	
	int i, n, s, lsd;
	struct sockaddr_in addr;
	socklen_t addrlen;
	u_short id;
	
	n = listen(lsd, 5);
	if(n < 0) {
		snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): listen()\n", __func__);
		ctx->error = 1;
		pthread_exit(NULL);
	}

	for(i = ctx->pid ; i < ctx->n ; i++) {
		s = accept(lsd, (struct sockaddr *)&addr, &addrlen);
		if(s < 0) {
			snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): accept()\n", __func__);
			ctx->error = 1;
			pthread_exit(NULL);
		}
		
		n = recv(s, &id, sizeof(id), 0); 
		if(s < 0) {
			snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): recv()\n", __func__);
			ctx->error = 1;
			pthread_exit(NULL);
		}
		ctx->group[i]->socket = s;
	}
	
	pthread_exit(NULL);
}
#endif

void ritas_reset_timer(ritas_t *ctx) {
	gettimeofday(&ctx->t_init, NULL);
}

long long ritas_get_timer(ritas_t *ctx) {
	struct timeval t1;
	long long dt;
	
	gettimeofday(&t1, NULL);
	//dt = ((t1.tv_sec*(long long)1000000)+t1.tv_usec) - ((ctx->t_init.tv_sec*(long long)1000000)+ctx->t_init.tv_usec);
	dt = toLL(t1)-toLL(ctx->t_init);

	return dt;
}
	
void ritas_register_decision_time(u_short pid, u_short cid, u_char decision, u_short round, struct timeval t0) {
	struct timeval t1;
	long long dt;
	
	gettimeofday(&t1, NULL);	
	dt = toLL(t1)-toLL(t0);
	printf("[%d] BC %d, decision time: %lld us, round: %d, value: %d\n", pid, cid, dt, round, decision);
}	
