#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <pthread.h>
#include <signal.h>
#include "ritas_mbuf.h"
#include "ritas_proc.h"
#include "ritas_netio.h"
#include "ritas_rc.h"
#include "ritas_rb.h"
#include "ritas_bc.h"
#include "ritas_mvc.h"
#include "ritas_vc.h"

int ritas_rc_send(ritas_t *ctx, ritas_rccb_t *rccb, ritas_mbuf_t *m, ritas_proc_t *proc) {
	return ritas_rc_output(ctx, rccb, m, proc);	
}

int ritas_rc_broadcast(ritas_t *ctx, ritas_rccb_t *rccb, ritas_mbuf_t *m) {
	return ritas_rc_output(ctx, rccb, m, NULL);
}

int ritas_rc_output(ritas_t *ctx, ritas_rccb_t *rccb, ritas_mbuf_t *m, ritas_proc_t *proc) {
	
	struct ritas_rc_header *rch;
#ifdef RELIABLE
	ritas_rbto_t *rbto;
#endif
	int i, n;

	/*
	 * Build RC Header
	 */
	m->head = m->head - RITAS_RC_HEADER_SIZE;
	if(m->head < m->data) {
	    snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): mbuf data overflow\n", __func__);
	    return -1;
	}
	rch = (struct ritas_rc_header *)m->head; 
	rch->id = rccb->id;
	rch->pid = ctx->pid;
	
	#ifdef DEBUG
	printf("[%d] RC(%d,%d,{%d,%d}) out", ctx->pid, rch->id.cid, rch->id.mvcid, rch->id.txid, rch->id.rbid);
	if(rch->id.flags & RITAS_ID_AB)
		printf(" ab");
	if(rch->id.flags & RITAS_ID_VC)
		printf(" vc");
	if(rch->id.flags & RITAS_ID_MVC)
		printf(" mvc");
	if(rch->id.flags & RITAS_ID_BC)
		printf(" bc");
	printf("\n");
	#endif

	if(proc) {
		n = ritas_send(ctx, proc, m);
		if(n < 0) {
			return -1;
		}
	}
	else {	
		for(i = 0 ; i < ctx->n ; i++) {
			n = ritas_send(ctx, ctx->group[i], m);
			if(n < 0) {
				return -1;
			}
		}
	}

#ifdef RELIABLE
	rbto = ritas_rbto_init(ctx, m);
	pthread_mutex_lock(&ctx->rbto_mutex);
	ritas_ht_add(ctx->rc_rbto_ht, (u_char *)&rch->mid, sizeof(rch->mid), rbto);
	pthread_mutex_unlock(&ctx->rbto_mutex);
	ritas_rbto_list_add(ctx, rbto);
#endif

	/*
	 * Destroy mbuf.
	 */
	ritas_mbuf_destroy(m);

	return 1;
}

int ritas_rc_input(ritas_t *ctx, ritas_mbuf_t *m) {

	ritas_rc_header_t *rch;
	ritas_list_t *mfifo;
	ritas_rccb_t *rccb;
#ifdef RELIABLE
	ritas_rbto_t *rbto;
#endif
	int r;

	rch = (ritas_rc_header_t *)m->head;

	#ifdef DEBUG
	printf("[%d] RC(%d,%d,{%d,%d}) in (%d bytes)", ctx->pid, rch->id.cid, rch->id.mvcid, rch->id.txid, rch->id.rbid, m->tail - m->head);
	if(rch->id.flags & RITAS_ID_AB)
		printf(" ab");
	if(rch->id.flags & RITAS_ID_VC)
		printf(" vc");
	if(rch->id.flags & RITAS_ID_MVC)
		printf(" mvc");
	if(rch->id.flags & RITAS_ID_BC)
		printf(" bc");
	printf(" from %d\n", rch->pid);
	#endif
	
	m->proc = ritas_proc_get(ctx, rch->pid);
	if(!m->proc) {
		ritas_mbuf_destroy(m);
	    return 1;
	}
	
	m->id = &rch->id;
	m->head = m->head + RITAS_RC_HEADER_SIZE;

#ifdef RELIABLE	
	/*
	 * Check if it's an ACK
	 */
	if(rch->type == RITAS_RC_ACK) {
		rbto = ritas_ht_lookup(ctx->rc_rbto_ht, (u_char *)&rbh->id, sizeof(rbh->id));
		if(!rbto) {
			ritas_mbuf_destroy(m);
			return 1;
		}
		if(rbto->proc[rbh->id]) {
			ritas_mbuf_destroy(m);
			return 1;
		}
		
		rbto->reply_count++;
		
		/*
		 * When reply count reaches 2f+1, update RTO.
		 */
		if(rbto->reply_count == ctx->n - ctx->f) {
			ritas_rc_elapsed(rbto->tx_time, &tv);
			ritas_rc_update_rto(ctx, &tv);
			ritas_rc_rtx(ctx);
		}
		/*
		 * When reply count reaches n we don't need the RBTO anymore.
		 */
		else if(rbto->reply_count == ctx->n) {
			pthread_mutex_lock(ctx->rbto_mutex);
			ritas_ht_remove(ctx->rc_rbto_ht, (u_char *)&rbh->id, sizeof(rbh->id));
			ritas_list_remove(ctx->rbto_l, rbto);
			pthread_mutex_unlock(ctx->rbto_mutex);
			
			ritas_rbto_destroy(rbto);
		}
	
		return 1;
	}
#endif
	
	/*
	 * Locate the control block.
	 */
	rccb = ritas_ht_lookup(ctx->ht_rccb, (u_char *)&rch->id, RITAS_ID_HEADER_SIZE);
	
	/*
	 * If we don't have a control block, then put the message in the
	 * "out-of-context" buffer.
	 */
	if(!rccb) {
		mfifo = ritas_ht_lookup(ctx->ht_ooc, (u_char *)&rch->id, RITAS_ID_HEADER_SIZE);
		if(!mfifo) {
			mfifo = ritas_list_init();
			if(!mfifo) {
			    snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): unable to initialize list\n", __func__);						
				return -1;
			}
			r = ritas_ht_add(ctx->ht_ooc, (u_char *)&rch->id, RITAS_ID_HEADER_SIZE, mfifo);
			
			if(r == -1) {
				snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): unable to insert object in hash table\n", __func__);
				return -1;
			}
			
		}
		r = ritas_list_enqueue(mfifo, m);
		if(r == -1) {
			snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): unable to insert object in hash table\n", __func__);
			return -1;
		}
		return 1;
	}

	if(rccb->parent_proto == RITAS_RC_PROTO_EB)
		r = ritas_eb_input(ctx, rccb->parent, m);
	else if(rccb->parent_proto == RITAS_RC_PROTO_RB)
		r = ritas_rb_input(ctx, rccb->parent, m);
	else
		r = ritas_bc_input(ctx, rccb->parent, m);

	return r;
}

ritas_rccb_t *ritas_rc_init(ritas_t *ctx, void *parent, u_char parent_proto) {
	
	ritas_rccb_t *rccb;
	int n;

	rccb = malloc(sizeof(*rccb));	
	if(!rccb) {
		snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): unable to malloc() control block\n", __func__);
		goto bad;
	}
	memset(rccb, '\0', sizeof(*rccb));
		
	rccb->parent = parent;
	rccb->parent_proto = parent_proto;

	/*
	 * Build RITAS ID
	 */
	ritas_rc_build_id(ctx, rccb);

	/*
	 * Add to OOC list.
	 */
	n = ritas_list_enqueue(ctx->l_ooc, rccb);
	if(n == -1) {
		snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): unable to add to list\n", __func__);
		goto bad;
	}
	ctx->scheduler_v = 1;

	/*
	 * Insert RCCB in the hash table
	 */
	n = ritas_ht_add(ctx->ht_rccb, (u_char *)&rccb->id, RITAS_ID_HEADER_SIZE, rccb);
	if(n == -1)
		goto bad;
		
	return rccb;

bad:
	if(rccb) {
		ritas_rc_destroy(ctx, rccb);
	}
	return NULL;
}

void ritas_rc_destroy(ritas_t *ctx, ritas_rccb_t *rccb) {

	ritas_list_t *mfifo;
	ritas_mbuf_t *m;
		
	/*
	 * Remove RCCB from the hash table.
	 */
	ritas_ht_remove(ctx->ht_rccb, (u_char *)&rccb->id, RITAS_ID_HEADER_SIZE);
	
	/*
	 * Remove RCCB from OOC list (if it's there in the first place)
	 */
	if(rccb->prev || rccb->next) {
		ritas_list_remove(ctx->l_ooc, rccb);
	}	

	/*
	 * Consume all OOC messages.
	 */
	if((mfifo = ritas_ht_remove(ctx->ht_ooc, (u_char *)&rccb->id, RITAS_ID_HEADER_SIZE))) {
		while((m = ritas_list_dequeue(mfifo)))
			ritas_mbuf_destroy(m);
		//ritas_list_destroy(mfifo);
	}
	ritas_list_remove(ctx->l_ooc, rccb);

	free(rccb);
}

int ritas_rc_renew(ritas_t *ctx, ritas_rccb_t *rccb, void *parent) {

	int n;
	ritas_mbuf_t *m;
	ritas_list_t *mfifo;
	
	if(!rccb) {
		return 1;
	}
	
	/*
	 * Remove RCCB from the hash table
	 */
	ritas_ht_remove(ctx->ht_rccb, (u_char *)&rccb->id, RITAS_ID_HEADER_SIZE);

	/*
	 * Remove RCCB from OOC list (if it's there in the first place)
	 */
	if(rccb->prev || rccb->next) {
		ritas_list_remove(ctx->l_ooc, rccb);
	}

	/*
	 * Consume all OOC messages.
	 */
	if((mfifo = ritas_ht_remove(ctx->ht_ooc, (u_char *)&rccb->id, RITAS_ID_HEADER_SIZE))) {
		while((m = ritas_list_dequeue(mfifo)))
			ritas_mbuf_destroy(m);
		//ritas_list_destroy(mfifo);
	}

	rccb->parent = parent;
	/*
	 * Build RITAS ID
	 */
	ritas_rc_build_id(ctx, rccb);
	
	/*
	 * Insert RCCB in the hash table
	 */
	n = ritas_ht_add(ctx->ht_rccb, (u_char *)&rccb->id, RITAS_ID_HEADER_SIZE, rccb);
	if(!n) {
		snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): unable to add to hash table\n", __func__);
		return -1;
	}		

	/*
	 * Add to OOC list.
	 */
	n = ritas_list_enqueue(ctx->l_ooc, rccb);
	if(n == -1) {
		snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): unable to add to list\n", __func__);
		return -1;
	}
	ctx->scheduler_v = 1;

	return 1;

}

void ritas_rc_build_id(ritas_t *ctx, ritas_rccb_t *rccb) {
	
	void *cb;
	u_char cb_type;
	
	/*
	 * Build RITAS ID
	 */
	rccb->id.flags = 0x00;
	rccb->id.mvcid = 0;
	rccb->id.cid = 0;

	if(rccb->parent_proto == RITAS_RC_PROTO_EB) {
		rccb->id.txid = ritas_eb_txid(rccb->parent);
		rccb->id.rbid = ritas_eb_rbid(rccb->parent);		
		cb = ritas_eb_parent(rccb->parent);
		cb_type = ritas_eb_parent_type(rccb->parent);
	}
	else if(rccb->parent_proto == RITAS_RC_PROTO_RB) {	
		rccb->id.txid = ritas_rb_txid(rccb->parent);
		rccb->id.rbid = ritas_rb_rbid(rccb->parent);	
		cb = ritas_rb_parent(rccb->parent);
		cb_type = ritas_rb_parent_type(rccb->parent);
	}
	else if(rccb->parent_proto == RITAS_RC_PROTO_BC) {
		cb_type = RITAS_BC;
		cb = rccb->parent;
		rccb->id.txid = ((ritas_bccb_t *)cb)->round; // Warning: shameless hammering!
		rccb->id.rbid = ((ritas_bccb_t *)cb)->step;	// Warning: shameless hammering!
	}
			
	if(cb) {
		switch(cb_type) {
		case RITAS_BC:
			rccb->id.cid = ritas_bc_id(cb);
			rccb->id.flags |= RITAS_ID_BC;
			cb = ritas_bc_parent(cb);
			if(cb) {
		case RITAS_MVC:
				rccb->id.mvcid = ritas_mvc_id(cb);
				rccb->id.flags |= RITAS_ID_MVC;
				cb = ritas_mvc_parent(cb);
				if(cb) {
		case RITAS_VC:
					rccb->id.cid = ritas_vc_id(cb);
					rccb->id.flags |= RITAS_ID_VC;
					cb = ritas_vc_parent(cb);
					if(cb) {
						rccb->id.flags |= RITAS_ID_AB;
						break;
		case RITAS_AB:
						rccb->id.cid = ritas_ab_id(cb);
						rccb->id.flags |= RITAS_ID_AB;
					}
				}
			}
		}
	}
}

/*
 * Timeout function. Blocks on sigwait() waiting for the current RTO to expire
 * via SIGALRM and calls retransmission routine.
 */
void ritas_rc_timeout(ritas_t *ctx) {
	
	int sig;
	sigset_t mask;
	
	sigemptyset(&mask);
	sigaddset(&mask, SIGALRM);
	pthread_sigmask(SIG_SETMASK, &mask, NULL);

	for(;;) {
		sigwait(&mask, &sig);
		ritas_rc_rtx(ctx);
		ritas_rc_reset_timer(ctx);
	}
}


void ritas_rc_reset_timer(ritas_t *ctx) {
	
	ritas_rbto_t *rbto;
	
	pthread_mutex_lock(&ctx->rbto_mutex);
	alarm(0);
	rbto = (ritas_rbto_t *)ctx->rbto_l->first;
	if(rbto) {
		ctx->itimer.it_value.tv_sec = (rbto->tx_time + ctx->rto)/1000000;
		ctx->itimer.it_value.tv_usec = (rbto->tx_time + ctx->rto)%1000000;
		setitimer(ITIMER_REAL, &ctx->itimer, NULL);
	}
	pthread_mutex_lock(&ctx->rbto_mutex);
	return;
}
	

int ritas_rc_rtx(ritas_t *ctx) {
	
	ritas_rbto_t *rbto;
	struct timeval tv;
	long long current_time;
	int i, n;
	
	gettimeofday(&tv, NULL);
	current_time = tv.tv_sec*(long long)1000000+tv.tv_usec;
	
	pthread_mutex_lock(&ctx->rbto_mutex);
	for(rbto = (ritas_rbto_t *)ctx->rbto_l->first ; rbto ; rbto = rbto->next) {
		if(rbto->tx_time >= current_time + ctx->rto) {
			for(i = 0 ; i < ctx->n ; i++) {
				if(!rbto->proc[i]) {
					n = ritas_send(ctx, ctx->group[i], rbto->m);
					if(n == -1)
						return -1;
				}
				ritas_list_remove(ctx->rbto_l, rbto);
				rbto->tx_time = current_time;
				ritas_list_enqueue(ctx->rbto_l, rbto);
			}
		}
		else {
			break;
		}
	}
	pthread_mutex_unlock(&ctx->rbto_mutex);

	return 1;
}
			

void ritas_rc_update_rto(ritas_t *ctx, struct timeval *measuredST) {
	
	long long delta;

	pthread_mutex_lock(&ctx->rbto_mutex);
	delta = (measuredST->tv_sec*(long long)1000000)+measuredST->tv_usec - ctx->srtt;
	ctx->srtt = ctx->srtt + (delta >> 3);
	ctx->rttvar = ctx->rttvar + ((delta - ctx->rttvar) >> 2);
	ctx->rto = ctx->srtt + 4*ctx->rttvar;
	pthread_mutex_unlock(&ctx->rbto_mutex);

	return;
}	

void ritas_rc_elapsed(long long t0, struct timeval *elapsed) {
	struct timeval tv;
	long long elap;
	
	gettimeofday(&tv, NULL);
	elap = tv.tv_sec*(long long)1000000+tv.tv_usec - t0;
	
	elapsed->tv_sec = elap/1000000;
	elapsed->tv_usec = elap%1000000;

	return;
}

ritas_rbto_t *ritas_rbto_init(ritas_t *ctx, ritas_mbuf_t *m) {
	
	ritas_rbto_t *rbto;
	struct timeval tv;
	
	rbto = malloc(sizeof(*rbto));
	if(!rbto) {
		snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): malloc()\n", __func__);
		goto bad;
	}

	rbto->proc = malloc(ctx->n*sizeof(rbto->proc));
	if(!rbto) {
		snprintf(ctx->errbuf, RITAS_ERRBUF_SIZE, "%s(): malloc()\n", __func__);
		goto bad;
	}
	
	rbto->prev = rbto->next = NULL;
	rbto->m = m;
	gettimeofday(&tv, NULL);
	rbto->tx_time = tv.tv_sec*(long long)1000000+tv.tv_usec;

bad:
	if(rbto)
		free(rbto);
		
	return NULL;
}

void ritas_rbto_destroy(ritas_rbto_t *rbto) {
	
	free(rbto->proc);
	ritas_mbuf_destroy(rbto->m);
	free(rbto);
}

int ritas_rbto_list_add(ritas_t *ctx, ritas_rbto_t *rbto) {

	pthread_mutex_lock(&ctx->rbto_mutex);

	ritas_list_enqueue(ctx->rbto_l, rbto);
	if(ctx->rbto_l->first == (void *)rbto)
		ritas_rc_reset_timer(ctx);

	pthread_mutex_unlock(&ctx->rbto_mutex);
	
	return 1;
}
	
int ritas_rbto_list_remove(ritas_t *ctx, ritas_rbto_t *rbto) {

	struct timeval tv;

	pthread_mutex_lock(&ctx->rbto_mutex);

	ritas_rc_elapsed(rbto->tx_time, &tv);
	
	ritas_list_remove(ctx->rbto_l, rbto);

	ritas_rc_update_rto(ctx, &tv);
	ritas_rc_rtx(ctx);
	ritas_rc_reset_timer(ctx);

	pthread_mutex_unlock(&ctx->rbto_mutex);
	
	return 1;
}
