	#include "ritas_thcoin.h"
#include "ritas_thcoin_keys.h"
#include <openssl/bn.h>
#include <openssl/sha.h>
#include <string.h>

ritas_thcoin_t *ritas_thcoin_init_bn(u_short id, u_short n, u_short k) {
	
	ritas_thcoin_t *coin;
	int i, r;
	BIGNUM **bns;
	
	/*
	 * Alloc main structure.
	 */
	coin = malloc(sizeof(*coin));
	if(!coin) {
		return NULL;
	}
	memset(coin, '\0', sizeof(*coin));

	
	coin->coin_name_s = 0;		// Reset coin name size.
	coin->id = id;				// Init process ID.
	coin->n = n;				// Init number of processes.
	coin->k = k;				// Init threshold processes

	coin->_g = BN_new();		// Create ~g
	
	bns = &coin->p;
	r = BN_dec2bn(bns, ritas_thcoin_p);			// Init p
	if(!r) {
		ritas_thcoin_destroy(coin);
		return NULL;
	}
	
	r = BN_dec2bn(&coin->q, ritas_thcoin_q);	// Init q
	if(!r) {
		ritas_thcoin_destroy(coin);
		return NULL;
	}
	
	r = BN_dec2bn(&coin->g, ritas_thcoin_g);	// Init g
	if(!r) {
		ritas_thcoin_destroy(coin);
		return NULL;
	}
		
	/*
	 * Init all public keys g_i, for i = 1 to n
	 */
	coin->gi = malloc(n*sizeof(BIGNUM *));
	if(!coin->gi) {
		ritas_thcoin_destroy(coin);
		return NULL;
	}
	memset(coin->gi, '\0', n*sizeof(BIGNUM *));
	for(i = 0 ; i < n ; i++) {
		r = BN_dec2bn(&coin->gi[i], ritas_thcoin_gi[i]);
		if(!r) {
			ritas_thcoin_destroy(coin);
			return NULL;
		}
	}

	r = BN_dec2bn(&coin->xi, ritas_thcoin_xi[id]);	// Init secret key x_i, i = id
	if(!r) {
		ritas_thcoin_destroy(coin);
		return NULL;
	}

	/*
	 * Reset individual shares
	 */
	coin->share = malloc(n*sizeof(u_char *));
	if(!coin->share) {
		ritas_thcoin_destroy(coin);
		return NULL;
	}
	memset(coin->share, '\0', n*sizeof(u_char *));
	
	return coin;
}


void ritas_thcoin_destroy(ritas_thcoin_t *coin) {
	
	int i;
	
	if(coin->share) {
		for(i = 0 ; i < coin->n ; i++)
			free(coin->share[i]);
		free(coin->share);
	}
	
	BN_free(coin->p);
	BN_free(coin->q);
	BN_free(coin->g);
	BN_free(coin->_g);
	for(i = 0 ; i < coin->n ; i++)
		BN_free(coin->gi[i]);
	free(coin->gi);
	free(coin->xi);
	
	free(coin);
}


int ritas_thcoin_reset(ritas_thcoin_t *coin, u_char *coin_name, u_short coin_name_s) {

	int repeat, r, i;
	u_char *buf;
	char ch;
	SHA_CTX sha_ctx;
	BN_CTX *bn_ctx;
	u_char coin_sha[SHA_DIGEST_LENGTH];
	BIGNUM *bnp, exp, *bn1;
	
	
	if(coin_name_s > RITAS_THCOIN_NAME_MAX)
		return -1;

	for(i = 0 ; i < coin->n ; i++)
		free(coin->share[i]);
	memset(coin->share, '\0', coin->n*sizeof(u_char *));		
		
	memcpy(coin->coin_name, coin_name, coin_name_s);
	coin->coin_name_s = coin_name_s;

	buf = NULL;
	bn_ctx = BN_CTX_new();
	if(!bn_ctx)
		goto bad;

	BN_init(&exp);

	/*************
	 * ~g = H(C) *
	 *************/
	/*
	 * Generate a hash from the name of the coin.
	 */
	SHA1(coin->coin_name, coin->coin_name_s, coin_sha);
	
	/*
	 * Expand the hash and convert it to a big number bn.
	 */
	repeat = (RITAS_THCOIN_MOD_SIZE + RITAS_THCOIN_KEY_SIZE) / (SHA_DIGEST_LENGTH * 8);
	buf = malloc(repeat * SHA_DIGEST_LENGTH + SHA_DIGEST_LENGTH);
	if(!buf)
		goto bad;
	memset(buf, '\0', repeat * SHA_DIGEST_LENGTH + SHA_DIGEST_LENGTH);
	for(ch = 0 ; ch < repeat ; ch++) {
		SHA1_Init(&sha_ctx);
		SHA1_Update(&sha_ctx, &ch, sizeof(ch)); 
		SHA1_Update(&sha_ctx, coin_sha, SHA_DIGEST_LENGTH);
		SHA1_Final(buf + ch*SHA_DIGEST_LENGTH, &sha_ctx);
	}
	memcpy(buf + repeat*SHA_DIGEST_LENGTH, coin_sha, SHA_DIGEST_LENGTH);
	
	bnp = BN_bin2bn(buf, repeat * SHA_DIGEST_LENGTH + SHA_DIGEST_LENGTH, coin->_g);
	if(!bnp)
		goto bad;

	/*
	 * ~g = bn^[(p-1)/q] mod p
	 */
	r = BN_sub(&exp, coin->p, BN_value_one());
	if(!r)
		goto bad;
		
	r = BN_div(&exp, NULL, &exp, coin->q, bn_ctx);
	if(!r)
		goto bad;

	r = BN_mod_exp(coin->_g, coin->_g, &exp, coin->p, bn_ctx); 
	if(!r)
		goto bad;

	BN_CTX_free(bn_ctx);	
	free(buf);
	return 1;
	
bad:
	if(bn_ctx)
		BN_CTX_free(bn_ctx);

	free(buf);
	
	return -1;
}


int ritas_thcoin_generate_share(ritas_thcoin_t *coin, u_char *share) {

	u_char coin_sha[SHA_DIGEST_LENGTH];
	u_char buf[RITAS_THCOIN_MOD_SIZE/8];
	int r, i;
	SHA_CTX sha_ctx;

	BN_CTX *bn_ctx;
	BIGNUM _g_i;
	BIGNUM exp, s, h, _h, c, z;

	bn_ctx = BN_CTX_new();
	if(!bn_ctx)
		goto bad;
		
	BN_init(&_g_i);
	BN_init(&exp);
	BN_init(&s);
	BN_init(&h);
	BN_init(&_h);
	BN_init(&c);
	BN_init(&z);


	/*****************
	 * ~g_i = ~g^x_i *	
	 *****************
	 */
	r = BN_mod_exp(&_g_i, coin->_g, coin->xi, coin->p, bn_ctx);
	if(!r)
		goto bad;
	
	/*
	 * s = random(Zq)
	 */
	r = BN_rand(&s, RITAS_THCOIN_KEY_SIZE, -1, 0);
	if(!r)
		goto bad;
	
	r = BN_mod(&s, &s, coin->q, bn_ctx);
	if(!r)
		goto bad;

	/*
	 * h = g^s
	 */
	r = BN_mod_exp(&h, coin->g, &s, coin->p, bn_ctx);
	if(!r)
		goto bad;
	
	/*
	 * ~h = ~g^s
	 */
	BN_mod_exp(&_h, coin->_g, &s, coin->p, bn_ctx);
	if(!r)
		goto bad;
	
	/*
	 * c = H'(g, g_i, h, ~g, ~g_i, ~h)
	 */
	SHA1_Init(&sha_ctx);
	r = BN_bn2bin(coin->g, buf);
	SHA1_Update(&sha_ctx, buf, r);
	r = BN_bn2bin(coin->gi[coin->id], buf);	// coin->gi[coin->id]
	SHA1_Update(&sha_ctx, buf, r);	 
	r = BN_bn2bin(&h, buf);
	SHA1_Update(&sha_ctx, buf, r);
	r = BN_bn2bin(coin->_g, buf);
	SHA1_Update(&sha_ctx, buf, r);
 	r = BN_bn2bin(&_g_i, buf);
	SHA1_Update(&sha_ctx, buf, r);
	r = BN_bn2bin(&_h, buf);
	SHA1_Update(&sha_ctx, buf, r);
	SHA1_Final(coin_sha, &sha_ctx);

	BN_bin2bn(coin_sha, SHA_DIGEST_LENGTH, &c);

	/*
	 * z = s + x_i*c
	 */
	r = BN_mod_mul(&z, coin->xi, &c, coin->q, bn_ctx);
	if(!r)
		goto bad;
	
	r = BN_mod_add(&z, &s, &z, coin->q, bn_ctx);
	if(!r)
		goto bad;

	/*
	 * The share is (~g_i, c, z)
	 */
	memset(share, '\0', 2*RITAS_THCOIN_KEY_SIZE_BYTES + RITAS_THCOIN_MOD_SIZE_BYTES);
	r = BN_num_bytes(&_g_i);
	BN_bn2bin(&_g_i, share + RITAS_THCOIN_MOD_SIZE_BYTES - r);
	r = BN_num_bytes(&c);
	BN_bn2bin(&c, share + RITAS_THCOIN_MOD_SIZE_BYTES + RITAS_THCOIN_KEY_SIZE_BYTES - r);
	r = BN_num_bytes(&z);
	BN_bn2bin(&z, share + RITAS_THCOIN_MOD_SIZE_BYTES + RITAS_THCOIN_KEY_SIZE_BYTES + RITAS_THCOIN_KEY_SIZE_BYTES - r);
	
	BN_CTX_free(bn_ctx);
	return 1;

bad:
	if(bn_ctx)
		BN_CTX_free(bn_ctx);
	
	return -1;
}


int ritas_thcoin_verify_share(ritas_thcoin_t *coin, u_char *share, u_short id) {

	BN_CTX *bn_ctx;
	SHA_CTX sha_ctx;
	BIGNUM gz, gc, h, _h, _g_j, c, z, c_;
	BIGNUM *bnp;
	u_char buf[RITAS_THCOIN_MOD_SIZE/8];
	u_char coin_sha[SHA_DIGEST_LENGTH];		
	int r;

	bn_ctx = BN_CTX_new();
	if(!bn_ctx)
		goto bad;

	BN_init(&gz);
	BN_init(&gc);
	BN_init(&_g_j);
	BN_init(&c);
	BN_init(&z);	
	BN_init(&h);
	BN_init(&_h);
	BN_init(&c_);
	
	BN_bin2bn(share, RITAS_THCOIN_MOD_SIZE/8, &_g_j);
	BN_bin2bn(share + RITAS_THCOIN_MOD_SIZE/8, RITAS_THCOIN_KEY_SIZE/8, &c);
	BN_bin2bn(share + RITAS_THCOIN_MOD_SIZE/8 + RITAS_THCOIN_KEY_SIZE/8, RITAS_THCOIN_KEY_SIZE/8, &z);	

	r = BN_mod_exp(&gz, coin->g, &z, coin->p, bn_ctx);
	if(!r) goto bad;

	r = BN_mod_exp(&gc, coin->gi[id], &c, coin->p, bn_ctx);
	if(!r) goto bad;
	
	bnp = BN_mod_inverse(&gc, &gc, coin->p, bn_ctx);
	if(!bnp) goto bad;
	
	r = BN_mod_mul(&h, &gz, &gc, coin->p, bn_ctx);
	if(!r) goto bad;

	r = BN_mod_exp(&gz, coin->_g, &z, coin->p, bn_ctx);
	if(!r) goto bad;
	
	r = BN_mod_exp(&gc, &_g_j, &c, coin->p, bn_ctx);
	if(!r) goto bad;
	
	bnp = BN_mod_inverse(&gc, &gc, coin->p, bn_ctx);
	if(!bnp) goto bad;
		
	r = BN_mod_mul(&_h, &gz, &gc, coin->p, bn_ctx);
	if(!r) goto bad;	

	// c = H'(g, g_i, h, ~g, ~g_i, ~h)
	SHA1_Init(&sha_ctx);
	r = BN_bn2bin(coin->g, buf);
	SHA1_Update(&sha_ctx, buf, r);
	r = BN_bn2bin(coin->gi[id], buf);
	SHA1_Update(&sha_ctx, buf, r);	 
	r = BN_bn2bin(&h, buf);
	SHA1_Update(&sha_ctx, buf, r);
	r = BN_bn2bin(coin->_g, buf);
	SHA1_Update(&sha_ctx, buf, r);
	r = BN_bn2bin(&_g_j, buf);
	SHA1_Update(&sha_ctx, buf, r);
	r = BN_bn2bin(&_h, buf);
	SHA1_Update(&sha_ctx, buf, r);
	SHA1_Final(coin_sha, &sha_ctx);

	BN_bin2bn(coin_sha, SHA_DIGEST_LENGTH, &c_);

	BN_CTX_free(bn_ctx);

	return !BN_cmp(&c, &c_) ? 1 : 0;


bad:
	if(bn_ctx)
		BN_CTX_free(bn_ctx);		

	return -1;
}

int ritas_thcoin_add_share(ritas_thcoin_t *coin, u_char *share, u_short id) {
	
	coin->share[id] = malloc(RITAS_THCOIN_SHARE_SIZE);
	if(!coin->share)
		return -1;
		
	memcpy(coin->share[id], share, RITAS_THCOIN_SHARE_SIZE);
	
	return 1;
}

int ritas_thcoin_combine_shares(ritas_thcoin_t *coin) {
	
	BN_CTX *bn_ctx;
	BIGNUM lambda;
	BIGNUM _gj, bni, bnj, bna, _g0;
	int i, j, count;

	bn_ctx = BN_CTX_new();
	if(!bn_ctx)
		goto bad;	
	BN_init(&lambda);	
	BN_init(&_gj);
	BN_init(&bni);
	BN_init(&bnj);
	BN_init(&bna);
	BN_init(&_g0);
	
	BN_one(&_g0);
	BN_zero(&bnj);
	for(j = 0, count = 0 ; j < coin->n && count < coin->k ; j++) {
		if(coin->share[j]) {
			/*
			 * Calculate (lamdba)_i,0 = PI(m/(m-i))
			 */			
			BN_one(&lambda);
			BN_zero(&bni);
			for(i = 0 ; i < coin->n ; i++) {
				if(coin->share[i] && i != j) {
					BN_add(&bna, &bni, BN_value_one());
					//BN_set_negative(&bna, 1);
					bna.neg = 1; // shiiiittt
					BN_mod_mul(&lambda, &lambda, &bna, coin->q, bn_ctx);
					BN_sub(&bna, &bnj, &bni);
					BN_mod_inverse(&bna, &bna, coin->q, bn_ctx);
					BN_mod_mul(&lambda, &lambda, &bna, coin->q, bn_ctx);
				}
				BN_add(&bni, &bni, BN_value_one());
			}

			/*
			 * Calculate ~g0
			 */	
			BN_bin2bn(coin->share[j], RITAS_THCOIN_MOD_SIZE/8, &_gj);		
			BN_mod_exp(&bna, &_gj, &lambda, coin->p, bn_ctx);
			BN_mod_mul(&_g0, &_g0, &bna, coin->p, bn_ctx);
			
			count++;
		}
		BN_add(&bnj, &bnj, BN_value_one());
	}
	if(count < coin->k)
		goto bad;

	//printf("_g0: %s\n", BN_bn2hex(&_g0));

	/*
	 * Hash ~g0 to domain {0,1} and we're done!
	 */
	i = BN_is_bit_set(&_g0, 0) ? 1 : 0;

	//debug
	//aux = NULL;
	//BN_dec2bn(&aux, ritas_thcoin_x0);
	//BN_mod_exp(&_g0, coin->_g, aux, coin->p, bn_ctx);
	//printf("real g0: %s\n", BN_bn2hex(&_g0));

	BN_CTX_free(bn_ctx);
	return i;
	
bad:
	if(bn_ctx)
		BN_CTX_free(bn_ctx);		

	return -1;
}
