#include <openssl/bn.h>
#include <openssl/sha.h>
#include <sys/types.h>
#include <string.h>
#include "ritas_vector.h"
#include "ritas_thsig.h"

ritas_thsig_t *ritas_thsig_init_ms(u_short mypid, u_short n) {
	FILE *fp;
	ritas_thsig_t *thsig;
#define RITAS_THSIG_FNAME_SIZE 32
	char fname[RITAS_THSIG_FNAME_SIZE];
	int i;
	
	/*
	 * Alloc main structure
	 */
	thsig = malloc(sizeof(*thsig));
	if(!thsig)
		return NULL;
	memset(thsig, '\0', sizeof(*thsig));

	thsig->id = mypid;	// set identifier
	thsig->n = n;	// set number of processes
	
	snprintf(fname, RITAS_THSIG_FNAME_SIZE, "rsaprivatekey%02d.der", mypid);
	fp = fopen(fname, "rb");
	if(!fp) {
		ritas_thsig_destroy(thsig);
		return NULL;
	}
	d2i_RSAPrivateKey_fp(fp, &thsig->private_key);
	fclose(fp);
	
	thsig->public_key = malloc(n*sizeof(*thsig->public_key));
	memset(thsig->public_key, '\0', n*sizeof(*thsig->public_key));
	for(i = 0 ; i < n ; i++) {
		snprintf(fname, RITAS_THSIG_FNAME_SIZE, "rsapublickey%02d.der", i);
		fp = fopen(fname, "rb");
		if(!fp) {
			ritas_thsig_destroy(thsig);
			return NULL;
		}
		d2i_RSAPublicKey_fp(fp, &thsig->public_key[i]);
		fclose(fp);		
	}

	thsig->data_s = 0;
	thsig->data = NULL;

	thsig->share = malloc(n*sizeof(*thsig->share));
	memset(thsig->share, '\0', n*sizeof(*thsig->share));
	thsig->share_count = 0;

	return thsig;
}

void ritas_thsig_destroy(ritas_thsig_t *thsig) {
	
	int i;
	
	RSA_free(thsig->private_key);
	for(i = 0 ; i < thsig->n ; i++) {
		RSA_free(thsig->public_key[i]);
	}
	free(thsig->share);
	free(thsig->public_key);

	free(thsig->data);
	
	free(thsig);
}

int ritas_thsig_reset(ritas_thsig_t *thsig, u_short data_s) {
	
	if(!thsig)
		return -1;
	
	if(data_s > thsig->data_s) {
		free(thsig->data);
		thsig->data = malloc(data_s);
		if(!thsig->data)
			return -1;
		thsig->data_s = data_s;
	}

	thsig->offset = 0;

	memset(thsig->share, '\0', thsig->n*sizeof(*thsig->share));
	thsig->share_count = 0;
	
	return 1;
}

int ritas_thsig_update(ritas_thsig_t *thsig, const u_char *data, u_int data_s) {

	if(!thsig)
		return -1;

	if(thsig->offset + data_s > thsig->data_s)
		return -1;

	memcpy(thsig->data + thsig->offset, data, data_s);

	thsig->offset += data_s;
		
	return 1;
}

int ritas_thsig_generate_share(ritas_thsig_t *thsig, u_char *share) {
	
	int r;
	u_int siglen;

	siglen = RITAS_THSIG_SHARE_SIZE;	
	r = RSA_sign(NID_sha1, thsig->data, thsig->offset, share, &siglen, thsig->private_key);
	if(!r)
		return -1;
	
	return siglen;
}

int ritas_thsig_verify_share(ritas_thsig_t *thsig, u_char *share, u_short id) {
	
	int r;
	
	r = RSA_verify(NID_sha1, thsig->data, thsig->offset, share, RITAS_THSIG_SHARE_SIZE, thsig->public_key[id]);
	
	return r;
}

int ritas_thsig_verify(ritas_thsig_t *thsig, u_char *sig, u_short sig_size, u_short k) {
	
	int i, correct;
	u_short *id;
	ritas_vector_t *vector;
	
	vector = ritas_vector_extract(sig, sig_size);
	
	correct = 0;
	for(i = 0 ; i < vector->size && correct < k ; i++) {
		id = (u_short *)vector->element[i];
		i++;
		correct += RSA_verify(NID_sha1, thsig->data, thsig->offset, vector->element[i], RITAS_THSIG_SHARE_SIZE, thsig->public_key[*id]);
	}
	
	ritas_vector_destroy(vector);
	
	return (correct >= k ? 1 : 0);
}

int ritas_thsig_add_share(ritas_thsig_t *thsig, u_char *share, u_short id) {
	
	if(!thsig->share[id]) {
		thsig->share[id] = share;
		thsig->share_count++;
	}

	return 1;
}

int ritas_thsig_combine_shares(ritas_thsig_t *thsig, u_short k, u_char *sig) {
	
	ritas_vector_t *vector;
	int i, count, index, sigsize;
	u_short *id_array;

	if(thsig->share_count < k)
		return -1;

	id_array = malloc(k*sizeof(*id_array));
	vector = ritas_vector_init(k*2);
	count = 0;
	index = 0;
	for(i = 0 ; i < thsig->n && vector->size < k*2 ; i++) {
		if(thsig->share[i]) {
			id_array[index] = i;
			ritas_vector_add(vector, (u_char *)&id_array[index], sizeof(*id_array));
			ritas_vector_add(vector, thsig->share[i], RITAS_THSIG_SHARE_SIZE);
			index++;			
		}
	}

	sigsize = vector->data_s;
	ritas_vector_build(vector, sig);

	ritas_vector_destroy(vector);
	free(id_array);
	
	return sigsize;
}

int ritas_thsig_sigsize(u_short k) {
	// THIS CODE IS POTENTIALLY UNRELIABLE
	return k*(RITAS_THSIG_SHARE_SIZE + sizeof(u_short))+ sizeof(u_short);
}
